#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import unittest
from mbed_tools.devices._internal.windows.windows_identifier import WindowsUID, is_device_instance_id
import random
import uuid
from tests.devices.markers import windows_only


def generateUID() -> WindowsUID:
    return WindowsUID(
        uid=str(uuid.uuid4()), raw_uid=f"/{uuid.uuid4()}&001", serial_number=f"SN{str(uuid.uuid4()).replace('-','')}"
    )


@windows_only
class TestWindowsUID(unittest.TestCase):
    def test_is_instance_id(self):
        # Testing that the values are likely to be instance IDs generated by the OS
        self.assertTrue(is_device_instance_id(None))
        self.assertTrue(is_device_instance_id("8&2F125EC6&0&0003"))
        self.assertTrue(is_device_instance_id("8&2f125ec6&0"))
        self.assertTrue(is_device_instance_id("8&2F125EC6&0&0002"))
        self.assertFalse(is_device_instance_id(""))
        self.assertFalse(is_device_instance_id("000440112138"))

    def test_uid_equality(self):
        uid1 = WindowsUID(uid="uid1", raw_uid="/uid1&001", serial_number=str(random.randint(1, 100)))
        # Equal testing with self and other types.
        self.assertNotEqual(uid1, None)
        self.assertIsNotNone(uid1)
        self.assertNotEqual(uid1, "")
        self.assertFalse(uid1 == "")
        self.assertFalse(uid1 == dict())
        self.assertEqual(uid1, uid1)
        self.assertTrue(uid1 == uid1)

        # Does not equal to completely different objects
        uid2 = WindowsUID(uid="uid2", raw_uid="/uid1&002", serial_number=None)
        self.assertNotEqual(uid1, uid2)

        # Equals other objects with same uid
        uid3 = WindowsUID(uid="uid1", raw_uid="/uid1&003", serial_number=str(random.randint(1, 100)))
        self.assertEqual(uid1, uid3)
        self.assertTrue(uid1 == uid3)

        # Equals other objects with similar uid (subset)
        uid5 = WindowsUID(uid="uid1&0114", raw_uid="/uid1&0114", serial_number=None)
        self.assertEqual(uid3, uid5)
        self.assertEqual(uid1, uid5)

        # Equals other objects with same serial number
        uid4 = WindowsUID(uid="uid4", raw_uid="/uid4&004", serial_number=uid1.serial_number)
        self.assertEqual(uid1, uid4)

        # Equals other objects with serial number same to uid
        uid6 = WindowsUID(uid="uid6454", raw_uid="/uid6454&006", serial_number="uid1")
        self.assertEqual(uid1, uid6)
        # Tests with real data examples: disk UIDs and equivalent USB hosts UIDs.
        # Daplink:
        uid7 = WindowsUID(
            uid="0240000034544e45001a00018aa900292011000097969900&0",
            raw_uid="0240000034544E45001A00018AA900292011000097969900&0",
            serial_number="0240000034544e45001a00018aa900292011000097969900",
        )
        uid8 = WindowsUID(
            uid="0240000034544e45001a00018aa900292011000097969900",
            raw_uid="0240000034544E45001A00018AA900292011000097969900",
            serial_number=None,
        )
        self.assertEqual(uid7, uid8)
        # JLink:
        uid9 = WindowsUID(
            uid="000440112138", raw_uid="9&DBDECF6&0&000440112138&0", serial_number="                       134657890"
        )
        uid10 = WindowsUID(uid="000440112138", raw_uid="000440112138", serial_number="8&2f125ec6&0")
        self.assertEqual(uid9, uid10)
        # STLink
        uid11 = WindowsUID(
            uid="0672ff574953867567051035",
            raw_uid="9&3849C7A8&0&0672FF574953867567051035&0",
            serial_number="0672FF574953867567051035",
        )
        uid12 = WindowsUID(
            uid="0672ff574953867567051035", raw_uid="0672FF574953867567051035", serial_number="8&254f12cf&0"
        )
        self.assertEqual(uid11, uid12)

    def test_serial_number(self):
        # Tests trying to determine the most plausible serial number from a set of values.
        uid1 = generateUID()
        self.assertEqual(uid1.presumed_serial_number, uid1.uid)
        self.assertNotEqual(uid1.presumed_serial_number, uid1.serial_number)
        uid2 = WindowsUID(uid="uid12&223", raw_uid="djfds;fj", serial_number=None)
        self.assertEqual(uid2.presumed_serial_number, uid2.uid)
        self.assertNotEqual(uid2.presumed_serial_number, uid2.serial_number)
        self.assertFalse(uid2.contains_genuine_serial_number())
        uid3 = WindowsUID(uid="uid12&223", raw_uid="djfds;fj", serial_number="12345679")
        self.assertNotEqual(uid3.presumed_serial_number, uid3.uid)
        self.assertEqual(uid3.presumed_serial_number, uid3.serial_number)
        self.assertTrue(uid3.contains_genuine_serial_number())

    def test_instanceid(self):
        # Tests trying to determine the most plausible instance IDs from a set of values.
        uid1 = generateUID()
        self.assertEqual(uid1.instance_id, uid1.serial_number)
        uid2 = WindowsUID(uid="uid12&223", raw_uid="djfds;fj", serial_number=None)
        self.assertEqual(uid2.instance_id, uid2.uid)
        uid3 = WindowsUID(uid="uid12&223", raw_uid="djfds;fj", serial_number="12345679")
        self.assertEqual(uid3.instance_id, uid3.uid)
        uid4 = WindowsUID(uid="12345687", raw_uid="djfds;fj", serial_number="12&3456&79")
        self.assertEqual(uid4.instance_id, uid4.serial_number)

    def test_uid_hashing(self):
        uid1 = generateUID()
        uid2 = generateUID()
        # Usual checks for different UIDs
        # Checks that if hashes are different then elements are not equal
        self.assertNotEqual(hash(uid1), hash(uid2))
        self.assertNotEqual(uid1, uid2)

        # Checks lookup in set
        self.assertIn(uid1, (uid1, uid2))
        self.assertIn(uid2, {uid1, uid2})
        # Checks lookup in dictionary
        self.assertIn(uid1, {uid1: "1", uid2: "2"})
        self.assertNotIn(uid1, dict())
        self.assertNotIn(uid1, {uid2: "1"})
        self.assertEqual({uid1: "1", uid2: "2"}.get(uid1), "1")
        self.assertIsNone({uid2: "2"}.get(uid1))

        # Checks the situation where two UIDs are equal/corresponding to the same device but their fields are different.
        uid3 = WindowsUID(uid="uid12&223", raw_uid="djfds;fj", serial_number=None)
        uid4 = WindowsUID(uid="123412", raw_uid="djfds;fjfsf&0&0000", serial_number="uid12&223")
        # UIDs being equal => hashes are equal.
        self.assertEqual(uid3, uid4)
        self.assertEqual(uid4, uid3)
        self.assertEqual(hash(uid3), hash(uid4))
        self.assertEqual(hash(uid4), hash(uid3))
        # Checks lookup in set
        self.assertIn(uid3, (uid1, uid4))
        self.assertIn(uid4, (uid1, uid3))

    def test_related_uid_lookup(self):
        # Tests dictionary lookup using real data.
        # UIDs are corresponding to Disk UIDs and related USB HUB UIDs.
        # Daplink
        uid11 = WindowsUID(
            uid="0240000034544e45001a00018aa900292011000097969900&0",
            raw_uid="0240000034544E45001A00018AA900292011000097969900&0",
            serial_number="0240000034544e45001a00018aa900292011000097969900",
        )
        uid12 = WindowsUID(
            uid="0240000034544e45001a00018aa900292011000097969900",
            raw_uid="0240000034544E45001A00018AA900292011000097969900",
            serial_number=None,
        )
        self.assertIn(uid12.presumed_serial_number, {uid11.presumed_serial_number: ""})
        # JLink
        uid21 = WindowsUID(
            uid="000440112138", raw_uid="9&DBDECF6&0&000440112138&0", serial_number="                       134657890"
        )
        uid22 = WindowsUID(uid="000440112138", raw_uid="000440112138", serial_number="8&2f125ec6&0")
        self.assertIn(uid21.presumed_serial_number, {uid22.presumed_serial_number: ""})
        # STLink
        uid31 = WindowsUID(
            uid="0672ff574953867567051035",
            raw_uid="9&3849C7A8&0&0672FF574953867567051035&0",
            serial_number="0672FF574953867567051035",
        )
        uid32 = WindowsUID(
            uid="0672ff574953867567051035", raw_uid="0672FF574953867567051035", serial_number="8&254f12cf&0"
        )
        self.assertIn(uid31.presumed_serial_number, {uid32.presumed_serial_number: ""})

    def test_ordering(self):
        uid1 = WindowsUID(uid="123456789", raw_uid="/uid1&002", serial_number=None)
        uid2 = WindowsUID(uid="0&64FAFGG", raw_uid="/0&64FAFGG&002", serial_number="345240562")
        self.assertGreater(uid2, uid1)
        self.assertLess(uid1, uid2)
        self.assertListEqual(sorted([uid2, uid1]), [uid1, uid2])



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase

from mbed_tools.build._internal.config.config import Config, Option, Macro
from tests.build._internal.config.factories import SourceFactory
from mbed_tools.build._internal.config.cumulative_data import CUMULATIVE_OVERRIDE_KEYS_IN_SOURCE
from mbed_tools.build._internal.config.bootloader_overrides import BOOTLOADER_OVERRIDE_KEYS_IN_SOURCE


class TestConfigFromSources(TestCase):
    def test_builds_config_from_sources(self):
        source_a = SourceFactory(config={"bool": True, "string": "foo"}, macros=["MACRO_A=A"])
        source_b = SourceFactory(config={"number": 1}, overrides={"bool": False}, macros=["MACRO_B=B"])

        config = Config.from_sources([source_a, source_b])

        self.assertEqual(
            config.options,
            {
                "bool": Option.build("bool", True, source_a).set_value(False, source_b),
                "number": Option.build("number", 1, source_b),
                "string": Option.build("string", "foo", source_a),
            },
        )

        self.assertEqual(
            config.macros,
            {"MACRO_A": Macro.build("MACRO_A=A", source_a), "MACRO_B": Macro.build("MACRO_B=B", source_b)},
        )

    def test_raises_when_trying_to_override_unset_option(self):
        source_a = SourceFactory(config={"bool": True})
        source_b = SourceFactory(overrides={"string": "hello"})

        with self.assertRaises(ValueError):
            Config.from_sources([source_a, source_b])

    def test_raises_when_trying_to_override_existing_macro(self):
        source_a = SourceFactory(macros=["MACRO_A=X"])
        source_b = SourceFactory(macros=["MACRO_A=Y"])

        with self.assertRaises(ValueError):
            Config.from_sources([source_a, source_b])

    def test_keeps_old_option_data(self):
        source_a = SourceFactory(config={"bool": {"help": "A simple bool", "value": True}})
        source_b = SourceFactory(overrides={"bool": False})

        config = Config.from_sources([source_a, source_b])

        self.assertEqual(config.options["bool"].help_text, "A simple bool")

    def test_does_not_explode_on_override_keys_used_by_other_parsers(self):
        for key in CUMULATIVE_OVERRIDE_KEYS_IN_SOURCE + BOOTLOADER_OVERRIDE_KEYS_IN_SOURCE:
            with self.subTest("Ignores override key '{key}'"):
                source_a = SourceFactory()
                source_b = SourceFactory(overrides={key: "boom?"})
                Config.from_sources([source_a, source_b])

    def test_ignores_present_option(self):
        source = SourceFactory(config={"mbed_component.present": {"help": "Mbed Component", "value": True}})

        config = Config.from_sources([source])

        self.assertFalse(config.options)


class TestOptionBuild(TestCase):
    def test_builds_option_from_config_data(self):
        source = SourceFactory()
        data = {
            "value": 123,
            "help": "some help text",
            "macro_name": "FOO_MACRO",
        }
        option = Option.build(key="target.stack-size", data=data, source=source)

        self.assertEqual(
            option,
            Option(
                key="target.stack-size",
                value=data["value"],
                help_text=data["help"],
                macro_name=data["macro_name"],
                set_by=source.human_name,
            ),
        )

    def test_sanitizes_booleans_into_ints(self):
        source = SourceFactory()
        option = Option.build(key="target.is-cool", data={"value": True}, source=source)
        self.assertEqual(option.value, 1)

        option = Option.build(key="target.is-cool", data=False, source=source)
        self.assertEqual(option.value, 0)

    def test_generates_macro_name_if_not_in_data(self):
        source = SourceFactory()
        data = {
            "value": 123,
            "help": "some help text",
        }
        option = Option.build(key="update-client.storage-size", data=data, source=source)

        self.assertEqual(option.macro_name, "MBED_CONF_UPDATE_CLIENT_STORAGE_SIZE")


class TestOptionSetValue(TestCase):
    def test_sanitizes_booleans_into_ints(self):
        option = Option.build(key="foo", data="some-data", source=SourceFactory())
        option.set_value(False, SourceFactory())

        self.assertEqual(option.value, 0)


class TestMacroBuild(TestCase):
    def test_builds_macros_with_value(self):
        source = SourceFactory()
        macro = Macro.build("FOO=BAR", source)

        self.assertEqual(
            macro, Macro(name="FOO", value="BAR", set_by=source.human_name),
        )

    def test_builds_macros_without_value(self):
        source = SourceFactory()
        macro = Macro.build("FOO", source)

        self.assertEqual(
            macro, Macro(name="FOO", value=None, set_by=source.human_name),
        )



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a Serial Port.

On Windows, Win32_SerialPort only represents physical serial Ports and hence, USB connections are not listed.
https://superuser.com/questions/835848/how-to-view-serial-com-ports-but-not-through-device-manager
https://stackoverflow.com/Questions/1388871/how-do-i-get-a-list-of-available-serial-ports-in-win32
https://stackoverflow.com/questions/1205383/listing-serial-com-ports-on-windows
https://serverfault.com/questions/398469/what-are-the-minimum-permissions-to-read-the-wmi-class-msserial-portname
"""

import re
from typing import NamedTuple, cast

from mbed_tools.devices._internal.windows.component_descriptor import (
    ComponentDescriptor,
    UNKNOWN_VALUE,
)

CAPTION_PATTERN = re.compile(r"^.* [(](.*)[)]$")


def parse_caption(caption: str) -> str:
    """Parses the caption string and returns the Port Name."""
    match = CAPTION_PATTERN.fullmatch(caption)
    return match.group(1) if match else UNKNOWN_VALUE


class PnPEntityMsdnDefinition(NamedTuple):
    """Msdn definition of a PnPEntity.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/win32-pnpentity
    """

    Availability: int
    Caption: str
    ClassGuid: str
    CompatibleID: list
    ConfigManagerErrorCode: int
    ConfigManagerUserConfig: bool
    CreationClassName: str
    Description: str
    DeviceID: str
    ErrorCleared: bool
    ErrorDescription: str
    HardwareID: list
    InstallDate: int
    LastErrorCode: int
    Manufacturer: str
    Name: str
    PNPClass: str
    PNPDeviceID: str
    PowerManagementCapabilities: int
    PowerManagementSupported: bool
    Present: bool
    Service: str
    Status: str
    StatusInfo: int
    SystemCreationClassName: str
    SystemName: str


class SerialPort(ComponentDescriptor):
    """Serial Port as defined in Windows API.

    As can be seen in Windows documentation,
    https://docs.microsoft.com/en-us/windows-hardware/drivers/install/system-defined-device-setup-classes-available-to-vendors#ports--com---lpt-ports--,
    ports are devices with ClassGuid = {4d36e978-e325-11ce-bfc1-08002be10318}. Hence the filter below.
    """

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(
            PnPEntityMsdnDefinition,
            win32_class_name="Win32_PnPEntity",
            win32_filter='ClassGuid="{4d36e978-e325-11ce-bfc1-08002be10318}"',
        )

    @property
    def component_id(self) -> str:
        """Returns the device id field."""
        return cast(str, self.get("DeviceID"))

    @property
    def port_name(self) -> str:
        """Gets the port name."""
        return parse_caption(self.get("Caption"))

    @property
    def pnp_id(self) -> str:
        """Gets the Plug and play id."""
        return cast(str, self.get("PNPDeviceID"))



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from dataclasses import fields
from unittest import TestCase

from mbed_tools.build._internal.config.bootloader_overrides import BootloaderOverrides, BootloaderOverride
from tests.build._internal.config.factories import SourceFactory


class TestCumulativeDataFromSources(TestCase):
    def test_assembles_metadata_from_sources(self):
        for field in fields(BootloaderOverrides):
            with self.subTest(f"Assemble {field.name}"):
                source = SourceFactory(overrides={f"target.{field.name}": ["FOO"]})

                bootloader_overrides = BootloaderOverrides.from_sources([source])

                self.assertEqual(
                    getattr(bootloader_overrides, field.name),
                    BootloaderOverride(set_by=source.human_name, name=field.name, value=["FOO"]),
                )



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a Volume Set.

CIM_VolumeSet should be the data model to use but does not seem to actually return the data that we are looking for:
https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/cim-volumeset
Therefore, a specific data model needs to be constructed using other Windows methods.
"""

from enum import Enum
from typing import NamedTuple, List
from mbed_tools.devices._internal.windows.component_descriptor import UNKNOWN_VALUE

import logging
import win32.win32api
import win32.win32file

logger = logging.getLogger(__name__)


class DriveType(Enum):
    """Drive type as defined in Win32 API.

    See https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-getdrivetypea.
    """

    DRIVE_UNKNOWN = 0  # The drive type cannot be determined.
    DRIVE_NO_ROOT_DIR = 1  # The root path is invalid; for example, there is no volume mounted at the specified path.
    DRIVE_REMOVABLE = (
        2
        # The drive has removable media; for example, a floppy drive, thumb drive, or flash card reader.
    )
    DRIVE_FIXED = 3  # The drive has fixed media; for example, a hard disk drive or flash drive.
    DRIVE_REMOTE = 4  # The drive is a remote (network) drive.
    DRIVE_CDROM = 5  # The drive is a CD-ROM drive.
    DRIVE_RAMDISK = 6  # The drive is a RAM disk.


class VolumeInformation(NamedTuple):
    """Volume information.

    See http://timgolden.me.uk/pywin32-docs/win32api__GetVolumeInformation_meth.html
    See also http://timgolden.me.uk/python/win32_how_do_i/find-drive-types.html
    """

    Name: str
    SerialNumber: int
    MaxComponentLengthOfAFileName: int
    SysFlags: int
    FileSystem: str
    UniqueName: str  # As defined by GetVolumeNameForVolumeMountPoint
    DriveType: DriveType  # As defined by GetDriveType


def _get_windows_volume_information(volume: str) -> List[str]:
    try:
        return list(win32.win32api.GetVolumeInformation(volume))
    except Exception as e:
        logger.debug(f"Cannot retrieve information about volume {volume}. Reason: {e}")
        return [UNKNOWN_VALUE] * 5


def _get_volume_name_for_mount_point(volume: str) -> str:
    try:
        return str(win32.win32file.GetVolumeNameForVolumeMountPoint(volume))
    except Exception as e:
        logger.debug(f"Cannot retrieve the real name of volume {volume}. Reason: {e}")
        return UNKNOWN_VALUE


def _get_drive_type(volume: str) -> DriveType:
    try:
        return DriveType(win32.win32file.GetDriveType(volume))
    except Exception as e:
        logger.debug(f"Cannot retrieve the type of volume {volume}. Reason: {e}")
        return DriveType.DRIVE_UNKNOWN


def get_volume_information(volume: str) -> VolumeInformation:
    """Gets the volume information."""
    if not volume.endswith("\\"):
        volume = f"{volume}\\"
    values: list = _get_windows_volume_information(volume) + [
        _get_volume_name_for_mount_point(volume),
        _get_drive_type(volume),  # type: ignore
    ]
    return VolumeInformation(*values)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a USB controller."""

from typing import NamedTuple, cast

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor


class UsbControllerMsdnDefinition(NamedTuple):
    """Msdn definition of a USB controller.

    See https://docs.microsoft.com/en-gb/windows/win32/cimwin32prov/win32-usbcontroller?redirectedfrom=MSDN
    Similar to https://docs.microsoft.com/en-gb/windows/win32/cimwin32prov/win32-usbcontrollerdevice
    """

    Availability: int
    Caption: str
    ConfigManagerErrorCode: int
    ConfigManagerUserConfig: bool
    CreationClassName: str
    Description: str
    DeviceID: str
    ErrorCleared: bool
    ErrorDescription: str
    InstallDate: int
    LastErrorCode: int
    Manufacturer: str
    MaxNumberControlled: int
    Name: str
    PNPDeviceID: str
    PowerManagementCapabilities: list
    PowerManagementSupported: bool
    ProtocolSupported: int
    Status: str
    StatusInfo: int
    SystemCreationClassName: str
    SystemName: str
    TimeOfLastReset: int


class UsbController(ComponentDescriptor):
    """USB Controller as defined in Windows API.

    See https://docs.microsoft.com/en-gb/windows/win32/cimwin32prov/win32-usbcontroller?redirectedfrom=MSDN
    Similar to https://docs.microsoft.com/en-gb/windows/win32/cimwin32prov/win32-usbcontrollerdevice
    """

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(UsbControllerMsdnDefinition, win32_class_name="Win32_USBController")

    @property
    def component_id(self) -> str:
        """Returns the device id field."""
        return cast(str, self.get("DeviceID"))



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from unittest import TestCase, mock

from click.testing import CliRunner

from mbed_tools.cli.project_management import new, clone, deploy, libs


@mock.patch("mbed_tools.cli.project_management.initialise_project", autospec=True)
class TestNewCommand(TestCase):
    def test_calls_new_function_with_correct_args(self, mock_initialise_project):
        CliRunner().invoke(new, ["path", "--create-only"])
        mock_initialise_project.assert_called_once_with(pathlib.Path("path"), True)

    def test_echos_mbed_os_message_when_required(self, mock_initialise_project):
        result = CliRunner().invoke(new, ["path"])

        self.assertEqual(
            result.output,
            "Creating a new Mbed program at path 'path'.\nDownloading mbed-os and adding it to the project.\n",
        )


@mock.patch("mbed_tools.cli.project_management.clone_project", autospec=True)
class TestCloneCommand(TestCase):
    def test_calls_clone_function_with_correct_args(self, mocked_clone_project):
        CliRunner().invoke(clone, ["url", "dst"])
        mocked_clone_project.assert_called_once_with("url", pathlib.Path("dst"), True)


@mock.patch("mbed_tools.cli.project_management.get_known_libs", autospec=True)
class TestLibsCommand(TestCase):
    def test_calls_libs_function(self, mocked_get_libs):
        CliRunner().invoke(libs)
        mocked_get_libs.assert_called_once()


@mock.patch("mbed_tools.cli.project_management.deploy_project", autospec=True)
class TestDeployCommand(TestCase):
    def test_calls_deploy_function_with_correct_args(self, mocked_deploy_project):
        CliRunner().invoke(deploy, ["path", "--force"])
        mocked_deploy_project.assert_called_once_with(pathlib.Path("path"), True)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import contextlib
import os
import pathlib

from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from click.testing import CliRunner

from mbed_tools.cli.build import build
from mbed_tools.project._internal.project_data import CMAKE_CONFIG_FILE_PATH, CMAKE_BUILD_DIR


@contextlib.contextmanager
def mock_project_directory(program, mbed_config_exists=False, build_tree_exists=False):
    with TemporaryDirectory() as tmp_dir:
        root = pathlib.Path(tmp_dir, "test-project")
        root.mkdir()
        program.root = root
        program.files.cmake_build_dir = root / CMAKE_BUILD_DIR
        program.files.cmake_config_file = root / CMAKE_CONFIG_FILE_PATH
        if mbed_config_exists:
            program.files.cmake_config_file.parent.mkdir(exist_ok=True)
            program.files.cmake_config_file.touch(exist_ok=True)

        if build_tree_exists:
            program.files.cmake_build_dir.mkdir(exist_ok=True)

        yield


@mock.patch("mbed_tools.cli.build.generate_build_system")
@mock.patch("mbed_tools.cli.build.build_project")
@mock.patch("mbed_tools.cli.build.MbedProgram")
@mock.patch("mbed_tools.cli.build.generate_config")
class TestBuildCommand(TestCase):
    def test_searches_for_mbed_program_at_default_project_path(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        runner = CliRunner()
        runner.invoke(build)

        mbed_program.from_existing.assert_called_once_with(pathlib.Path(os.getcwd()))

    def test_searches_for_mbed_program_at_user_defined_project_root(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        project_path = "my-blinky"

        runner = CliRunner()
        runner.invoke(build, ["-p", project_path])

        mbed_program.from_existing.assert_called_once_with(pathlib.Path(project_path))

    def test_calls_generate_build_system_if_build_tree_nonexistent(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=False):
            runner = CliRunner()
            runner.invoke(build, ["-m", "k64f", "-t", "gcc_arm"])

            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")

    def test_generate_build_system_not_called_if_build_tree_exists(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
            runner = CliRunner()
            runner.invoke(build)

            generate_build_system.assert_not_called()

    def test_generate_config_called_if_config_script_nonexistent(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=False):
            target = "k64f"
            toolchain = "gcc_arm"

            runner = CliRunner()
            runner.invoke(build, ["-t", toolchain, "-m", target])

            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)

    def test_raises_if_gen_config_toolchain_not_passed_when_required(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=False):
            target = "k64f"

            runner = CliRunner()
            result = runner.invoke(build, ["-m", target])

            self.assertIsNotNone(result.exception)
            self.assertRegex(result.output, "--toolchain")

    def test_raises_if_gen_config_target_not_passed_when_required(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=False):
            toolchain = "gcc_arm"

            runner = CliRunner()
            result = runner.invoke(build, ["-t", toolchain])

            self.assertIsNotNone(result.exception)
            self.assertRegex(result.output, "--mbed-target")

    def test_build_system_regenerated_when_target_and_toolchain_passed(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
            toolchain = "gcc_arm"
            target = "k64f"

            runner = CliRunner()
            runner.invoke(build, ["-t", toolchain, "-m", target])

            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")

    def test_build_folder_removed_when_clean_flag_passed(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
            toolchain = "gcc_arm"
            target = "k64f"

            runner = CliRunner()
            runner.invoke(build, ["-t", toolchain, "-m", target, "-c"])

            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")
            self.assertFalse(program.files.cmake_build_dir.exists())



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for project_data.py."""
import pathlib

from unittest import TestCase, mock

from mbed_tools.project._internal.project_data import (
    MbedProgramFiles,
    MbedOS,
    MAIN_CPP_FILE_NAME,
)
from tests.project.factories import make_mbed_lib_reference, make_mbed_program_files, make_mbed_os_files, patchfs


class TestMbedProgramFiles(TestCase):
    @patchfs
    def test_from_new_raises_if_program_files_already_exist(self, fs):
        root = pathlib.Path(fs, "foo")
        make_mbed_program_files(root)

        with self.assertRaises(ValueError):
            MbedProgramFiles.from_new(root)

    @patchfs
    def test_from_new_returns_valid_program_file_set(self, fs):
        root = pathlib.Path(fs, "foo")
        root.mkdir()

        program = MbedProgramFiles.from_new(root)

        self.assertTrue(program.app_config_file.exists())
        self.assertTrue(program.mbed_os_ref.exists())
        self.assertTrue(program.cmakelists_file.exists())

    @patchfs
    def test_from_new_calls_render_template_for_gitignore_cmake_and_main(self, fs):
        with mock.patch(
            "mbed_tools.project._internal.project_data.render_cmakelists_template"
        ) as render_cmakelists_template:
            with mock.patch(
                "mbed_tools.project._internal.project_data.render_main_cpp_template"
            ) as render_main_cpp_template:
                with mock.patch(
                    "mbed_tools.project._internal.project_data.render_gitignore_template"
                ) as render_gitignore_template:
                    root = pathlib.Path(fs, "foo")
                    root.mkdir()
                    program_files = MbedProgramFiles.from_new(root)
                    render_cmakelists_template.assert_called_once_with(program_files.cmakelists_file, "foo")
                    render_main_cpp_template.assert_called_once_with(root / MAIN_CPP_FILE_NAME)
                    render_gitignore_template.assert_called_once_with(root / ".gitignore")

    @patchfs
    def test_from_existing_finds_existing_program_data(self, fs):
        root = pathlib.Path(fs, "foo")
        make_mbed_program_files(root)

        program = MbedProgramFiles.from_existing(root)

        self.assertTrue(program.app_config_file.exists())
        self.assertTrue(program.mbed_os_ref.exists())
        self.assertTrue(program.cmakelists_file.exists())


class TestMbedLibReference(TestCase):
    @patchfs
    def test_is_resolved_returns_true_if_source_code_dir_exists(self, fs):
        root = pathlib.Path(fs, "foo")
        lib = make_mbed_lib_reference(root, resolved=True)

        self.assertTrue(lib.is_resolved())

    @patchfs
    def test_is_resolved_returns_false_if_source_code_dir_doesnt_exist(self, fs):
        root = pathlib.Path(fs, "foo")
        lib = make_mbed_lib_reference(root)

        self.assertFalse(lib.is_resolved())

    @patchfs
    def test_get_git_reference_returns_lib_file_contents(self, fs):
        root = pathlib.Path(fs, "foo")
        url = "https://github.com/mylibrepo"
        ref = "latest"
        references = [f"{url}#{ref}", f"{url}/#{ref}"]

        for full_ref in references:
            lib = make_mbed_lib_reference(root, ref_url=full_ref)

            reference = lib.get_git_reference()

            self.assertEqual(reference.repo_url, url)
            self.assertEqual(reference.ref, ref)


class TestMbedOS(TestCase):
    @patchfs
    def test_from_existing_finds_existing_mbed_os_data(self, fs):
        root_path = pathlib.Path(fs, "my-version-of-mbed-os")
        make_mbed_os_files(root_path)

        mbed_os = MbedOS.from_existing(root_path)

        self.assertEqual(mbed_os.targets_json_file, root_path / "targets" / "targets.json")

    @patchfs
    def test_raises_if_files_missing(self, fs):
        root_path = pathlib.Path(fs, "my-version-of-mbed-os")
        root_path.mkdir()

        with self.assertRaises(ValueError):
            MbedOS.from_existing(root_path)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Loads System's USB hub."""

from typing import Dict, List, cast, Optional, Set, Generator

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor
from mbed_tools.devices._internal.windows.device_instance_id import get_children_instance_id
from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader, ComponentsLoader
from mbed_tools.devices._internal.windows.usb_controller import UsbController
from mbed_tools.devices._internal.windows.usb_device_identifier import parse_device_id, UsbIdentifier
from mbed_tools.devices._internal.windows.usb_hub import UsbHub


class SystemUsbDeviceInformation:
    """Usb Hub cache for this system.

    On Windows, each interface e.g. Composite, Mass storage, Port is defined as
    a separate independent UsbHub although they are related to the same device.
    This cache tries to reduce the list of UsbHubs to only genuinely different devices.
    """

    def __init__(self, data_loader: SystemDataLoader) -> None:
        """Initialiser."""
        self._cache: Optional[Dict[UsbIdentifier, List[UsbHub]]] = None
        self._ids_cache: Optional[Set[UsbIdentifier]] = None
        self._data_loader = data_loader

    def _list_usb_controller_ids(self) -> List[UsbIdentifier]:
        return cast(
            List[UsbIdentifier],
            [
                parse_device_id(cast(UsbController, usbc).component_id)
                for usbc in ComponentsLoader(self._data_loader, UsbController).element_generator()
            ],
        )

    def _iterate_over_hubs(self) -> Generator[ComponentDescriptor, None, None]:
        return ComponentsLoader(self._data_loader, UsbHub).element_generator()

    def _populate_id_cache_with_non_serialnumbers(self) -> None:
        if not self._cache or not self._ids_cache:
            return
        for usb_id in self._cache:
            if usb_id not in self._ids_cache:
                self._ids_cache.add(usb_id)

    def _determine_potential_serial_number(self, usb_device: UsbHub) -> Optional[str]:
        return get_children_instance_id(usb_device.pnp_id)

    def _load(self) -> None:
        """Populates the cache."""
        self._cache = cast(Dict[UsbIdentifier, List[UsbHub]], dict())
        self._ids_cache = cast(Set[UsbIdentifier], set())
        controllers = self._list_usb_controller_ids()
        for usb_device in self._iterate_over_hubs():
            usb_id = parse_device_id(
                usb_device.component_id, serial_number=self._determine_potential_serial_number(cast(UsbHub, usb_device))
            )
            if usb_id in controllers:
                continue
            entry = self._cache.get(usb_id, list())
            entry.append(cast(UsbHub, usb_device))
            self._cache[usb_id] = entry

            if usb_id.contains_genuine_serial_number():
                self._ids_cache.add(usb_id)
        self._populate_id_cache_with_non_serialnumbers()

    @property
    def usb_devices(self) -> Dict[UsbIdentifier, List[UsbHub]]:
        """Usb devices present in the system."""
        if not self._cache:
            self._load()
        return cast(Dict[UsbIdentifier, List[UsbHub]], self._cache)

    def get_usb_devices(self, uid: UsbIdentifier) -> List[UsbHub]:
        """Gets all USB devices related to an identifier."""
        return self.usb_devices.get(uid, list())

    def usb_device_ids(self) -> List[UsbIdentifier]:
        """Gets system usb device IDs."""
        if not self._ids_cache:
            self._load()
        return cast(List[UsbIdentifier], self._ids_cache)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Ability to parse cumulative attributes from Sources."""
import itertools
import re
from dataclasses import dataclass, field, fields
from typing import Any, Iterable, Set, Tuple

from mbed_tools.build._internal.config.source import Source


@dataclass
class CumulativeData:
    """Representation of cumulative attributes assembled during Source parsing."""

    features: Set[str] = field(default_factory=set)
    components: Set[str] = field(default_factory=set)
    labels: Set[str] = field(default_factory=set)
    extra_labels: Set[str] = field(default_factory=set)
    device_has: Set[str] = field(default_factory=set)
    macros: Set[str] = field(default_factory=set)

    @classmethod
    def from_sources(cls, sources: Iterable[Source]) -> "CumulativeData":
        """Interrogate each Source in turn to create final CumulativeData."""
        data = CumulativeData()
        for source in sources:
            for key, value in source.overrides.items():
                if key in CUMULATIVE_OVERRIDE_KEYS_IN_SOURCE:
                    _modify_field(data, key, value)
        return data


def _modify_field(data: CumulativeData, key: str, value: Any) -> None:
    """Mutates CumulativeData in place by adding, removing or resetting the value of a field."""
    key, modifier = _extract_target_modifier_data(key)
    if modifier == "add":
        new_value = getattr(data, key) | set(value)
    elif modifier == "remove":
        new_value = getattr(data, key) - set(value)
    else:
        new_value = set(value)
    setattr(data, key, new_value)


_CUMULATIVE_FIELDS = [f.name for f in fields(CumulativeData)]
_PREFIXED_CUMULATIVE_FIELDS = [f"target.{f}" for f in _CUMULATIVE_FIELDS]
CUMULATIVE_OVERRIDE_KEYS_IN_SOURCE = _PREFIXED_CUMULATIVE_FIELDS + [
    f"{attr}_{suffix}" for attr, suffix in itertools.product(_PREFIXED_CUMULATIVE_FIELDS, ["add", "remove"])
]


def _extract_target_modifier_data(key: str) -> Tuple[str, str]:
    regex = fr"""
            (?P<key>{'|'.join(_CUMULATIVE_FIELDS)}) # attribute name (one of ACCUMULATING_OVERRIDES)
            _?                                      # separator
            (?P<modifier>(add|remove)?)             # modifier (add, remove or empty)
    """
    match = re.search(regex, key, re.VERBOSE)
    if not match:
        raise ValueError(f"Not a target modifier key {key}")
    return match["key"], match["modifier"]



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from functools import wraps
from tempfile import TemporaryDirectory
from mbed_tools.project._internal.libraries import MbedLibReference
from mbed_tools.project._internal.project_data import (
    CMAKELISTS_FILE_NAME,
    APP_CONFIG_FILE_NAME,
    MBED_OS_REFERENCE_FILE_NAME,
)


def patchfs(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        with TemporaryDirectory() as fs:
            func(*args, fs=fs, **kwargs)

    return wrapper


def make_mbed_program_files(root, config_file_name=APP_CONFIG_FILE_NAME):
    if not root.exists():
        root.mkdir()

    (root / MBED_OS_REFERENCE_FILE_NAME).touch()
    (root / config_file_name).touch()
    (root / CMAKELISTS_FILE_NAME).touch()


def make_mbed_lib_reference(root, name="mylib.lib", resolved=False, ref_url=None):
    ref_file = root / name
    source_dir = ref_file.with_suffix("")
    if not root.exists():
        root.mkdir()

    ref_file.touch()

    if resolved:
        source_dir.mkdir()

    if ref_url is not None:
        ref_file.write_text(ref_url)

    return MbedLibReference(reference_file=ref_file, source_code_path=source_dir)


def make_mbed_os_files(root):
    if not root.exists():
        root.mkdir()

    targets_dir = root / "targets"
    targets_dir.mkdir()
    (targets_dir / "targets.json").touch()



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import unittest
from mbed_tools.devices._internal.windows.component_descriptor_utils import (
    is_undefined_data_object,
    is_undefined_value,
    data_object_to_dict,
    UNKNOWN_VALUE,
)
from collections import namedtuple
import random


def generate_valid_values():
    return random.choice(["a test", 4646.454, 54, True])


def generate_undefined_values():
    return random.choice([0, None, False, UNKNOWN_VALUE])


class TestUtilities(unittest.TestCase):
    def test_is_value_undefined(self):
        self.assertTrue(is_undefined_value(generate_undefined_values()))
        self.assertFalse(is_undefined_value(generate_valid_values()))

    def test_is_data_object_undefined(self):
        field_number = 30
        DataObjectType = namedtuple("data_object_example", [f"field{i}" for i in range(0, field_number)])
        test1 = DataObjectType(*[generate_undefined_values() for i in range(0, field_number)])
        self.assertTrue(is_undefined_data_object(test1))
        test2 = DataObjectType(*[generate_valid_values() for i in range(0, field_number)])
        self.assertFalse(is_undefined_data_object(test2))

    def test_to_dict(self):
        field_number = 30
        DataObjectType = namedtuple("data_object_example", [f"field{i}" for i in range(0, field_number)])
        expected_dictionary = {
            f"field{i}": random.choice([generate_valid_values(), generate_undefined_values()])
            for i in range(0, field_number)
        }
        test = DataObjectType(**expected_dictionary)
        self.assertDictEqual(data_object_to_dict(test), expected_dictionary)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Test Linux Device Detector."""

from collections import namedtuple
from unittest import TestCase, mock, skipIf
from mbed_tools.devices._internal.candidate_device import CandidateDevice

try:
    from mbed_tools.devices._internal.linux import device_detector

    import_succeeded = True
except ImportError:
    import_succeeded = False


def mock_device_factory(**props):
    return namedtuple("MockDevice", "properties")(props)


@skipIf(not import_succeeded, "Tests require package dependencies only used on Linux.")
class TestLinuxDeviceDetector(TestCase):
    @mock.patch("mbed_tools.devices._internal.linux.device_detector.pyudev.Context")
    @mock.patch("mbed_tools.devices._internal.linux.device_detector._find_fs_mounts_for_device")
    @mock.patch("mbed_tools.devices._internal.linux.device_detector._find_serial_port_for_device")
    def test_builds_list_of_candidates(self, mock_find_serial_port, mock_find_fs_mounts, mock_udev_context):
        expected_serial = "2090290209"
        expected_vid = "0x45"
        expected_pid = "0x48"
        expected_fs_mount = ["/media/user/DAPLINK"]
        mock_find_serial_port.return_value = None
        mock_find_fs_mounts.return_value = expected_fs_mount
        devs = [
            mock_device_factory(
                ID_SERIAL_SHORT=expected_serial,
                ID_VENDOR_ID=expected_vid,
                ID_MODEL_ID=expected_pid,
                DEVNAME="/dev/sdabcde",
            )
        ]
        mock_udev_context().list_devices.return_value = devs
        detector = device_detector.LinuxDeviceDetector()
        candidates = detector.find_candidates()
        self.assertEqual(
            candidates,
            [
                CandidateDevice(
                    serial_number=expected_serial,
                    vendor_id=expected_vid,
                    product_id=expected_pid,
                    mount_points=expected_fs_mount,
                )
            ],
        )

    @mock.patch("mbed_tools.devices._internal.linux.device_detector.pyudev.Context")
    @mock.patch("mbed_tools.devices._internal.linux.device_detector._find_fs_mounts_for_device")
    def test_handles_filesystem_mountpoint_error_and_skips_device(self, mock_find_fs_mounts, mock_udev_context):
        mock_find_fs_mounts.return_value = []
        devs = [
            mock_device_factory(
                ID_SERIAL_SHORT="2090290209", ID_VENDOR_ID="0x45", ID_MODEL_ID="0x48", DEVNAME="/dev/sdabcde",
            )
        ]
        mock_udev_context().list_devices.return_value = devs
        detector = device_detector.LinuxDeviceDetector()
        candidates = detector.find_candidates()
        self.assertEqual(candidates, [])

    @mock.patch("mbed_tools.devices._internal.linux.device_detector.pyudev.Context")
    def test_finds_serial_port_with_matching_serial_id(self, mock_udev_context):
        matching_serial = "a"
        disk_device = mock_device_factory(ID_SERIAL_SHORT=matching_serial, DEVNAME="/dev/sdc")
        serial_device_match = mock_device_factory(ID_SERIAL_SHORT=matching_serial, DEVNAME="/dev/ttyACM0")
        serial_device_diff = mock_device_factory(ID_SERIAL_SHORT="b", DEVNAME="/dev/ttyUSB0")
        mock_udev_context().list_devices.return_value = [serial_device_match, serial_device_diff]
        serial_port = device_detector._find_serial_port_for_device(disk_device.properties["ID_SERIAL_SHORT"])
        self.assertEqual(serial_port, serial_device_match.properties["DEVNAME"])

    @mock.patch("mbed_tools.devices._internal.linux.device_detector.pyudev.Context")
    def test_returns_empty_none_when_no_matching_serial_id(self, mock_udev_context):
        disk_device = mock_device_factory(ID_SERIAL_SHORT="i", DEVNAME="/dev/sdc")
        serial_device = mock_device_factory(ID_SERIAL_SHORT="a", DEVNAME="/dev/ttyACM0")
        mock_udev_context().list_devices.return_value = [serial_device]
        serial_port = device_detector._find_serial_port_for_device(disk_device.properties["ID_SERIAL_SHORT"])
        self.assertEqual(serial_port, None)

    @mock.patch("mbed_tools.devices._internal.linux.device_detector.psutil")
    def test_finds_fs_mountpoints_for_device_files(self, mock_psutil):
        partition = namedtuple("Partition", "mountpoint,device")("/media/user/DAPLINK", "/dev/sdc")
        mock_psutil.disk_partitions.return_value = [partition]
        mounts = device_detector._find_fs_mounts_for_device("/dev/sdc")
        self.assertEqual(str(mounts[0]), partition.mountpoint)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Exceptions internal to the package."""

from mbed_tools.lib.exceptions import ToolsError


class SystemException(ToolsError):
    """Exception with regards to the underlying operating system."""


class NoBoardForCandidate(ToolsError):
    """Raised when board data cannot be determined for a candidate."""



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import factory
import pathlib

from mbed_tools.devices._internal.candidate_device import CandidateDevice


class CandidateDeviceFactory(factory.Factory):
    class Meta:
        model = CandidateDevice

    product_id = factory.Faker("hexify")
    vendor_id = factory.Faker("hexify")
    mount_points = [pathlib.Path(".")]
    serial_number = factory.Faker("hexify", text=("^" * 20))  # 20 characters serial number
    serial_port = None



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from mbed_tools.project._internal.render_templates import (
    render_cmakelists_template,
    render_main_cpp_template,
    render_gitignore_template,
)


@mock.patch("mbed_tools.project._internal.render_templates.datetime")
class TestRenderTemplates(TestCase):
    def test_renders_cmakelists_template(self, mock_datetime):
        with TemporaryDirectory() as tmpdir:
            the_year = 3999
            mock_datetime.datetime.now.return_value.year = the_year
            program_name = "mytestprogram"
            file_path = Path(tmpdir, "mytestpath")

            render_cmakelists_template(file_path, program_name)

            output = file_path.read_text()
            self.assertIn(str(the_year), output)
            self.assertIn(program_name, output)

    def test_renders_main_cpp_template(self, mock_datetime):
        with TemporaryDirectory() as tmpdir:
            the_year = 3999
            mock_datetime.datetime.now.return_value.year = the_year
            file_path = Path(tmpdir, "mytestpath")

            render_main_cpp_template(file_path)

            self.assertIn(str(the_year), file_path.read_text())

    def test_renders_gitignore_template(self, _):
        with TemporaryDirectory() as tmpdir:
            file_path = Path(tmpdir, "mytestpath")

            render_gitignore_template(file_path)

            self.assertIn("cmake_build", file_path.read_text())
            self.assertIn(".mbedbuild", file_path.read_text())



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase

from tests.build._internal.config.factories import ConfigFactory
from mbed_tools.build._internal.cmake_file import generate_mbed_config_cmake_file, _render_mbed_config_cmake_template


class TestGenerateCMakeListsFile(TestCase):
    def test_correct_arguments_passed(self):
        target = dict()
        target["labels"] = ["foo"]
        target["extra_labels"] = ["morefoo"]
        target["features"] = ["bar"]
        target["components"] = ["baz"]
        target["macros"] = ["macbaz"]
        target["device_has"] = ["stuff"]
        target["c_lib"] = ["c_lib"]
        target["core"] = ["core"]
        target["printf_lib"] = ["printf_lib"]
        target["supported_form_factors"] = ["arduino"]
        config = ConfigFactory()
        mbed_target = "K64F"
        toolchain_name = "GCC"
        target["supported_c_libs"] = {toolchain_name.lower(): ["small", "std"]}
        target["supported_application_profiles"] = ["full", "bare-metal"]

        result = generate_mbed_config_cmake_file(mbed_target, target, config, toolchain_name)

        self.assertEqual(
            result, _render_mbed_config_cmake_template(target, config, toolchain_name, mbed_target,),
        )


class TestRendersCMakeListsFile(TestCase):
    def test_returns_rendered_content(self):
        target = dict()
        target["labels"] = ["foo"]
        target["extra_labels"] = ["morefoo"]
        target["features"] = ["bar"]
        target["components"] = ["baz"]
        target["macros"] = ["macbaz"]
        target["device_has"] = ["stuff"]
        target["core"] = ["core"]
        target["c_lib"] = ["c_lib"]
        target["printf_lib"] = ["printf_lib"]
        target["supported_form_factors"] = ["arduino"]
        config = ConfigFactory()
        toolchain_name = "baz"
        target["supported_c_libs"] = {toolchain_name.lower(): ["small", "std"]}
        target["supported_application_profiles"] = ["full", "bare-metal"]
        result = _render_mbed_config_cmake_template(target, config, toolchain_name, "target_name")

        for label in target["labels"] + target["extra_labels"]:
            self.assertIn(label, result)

        for macro in target["features"] + target["components"] + [toolchain_name]:
            self.assertIn(macro, result)

        for toolchain in target["supported_c_libs"]:
            self.assertIn(toolchain, result)
            for supported_c_libs in toolchain:
                self.assertIn(supported_c_libs, result)

        for supported_application_profiles in target["supported_application_profiles"]:
            self.assertIn(supported_application_profiles, result)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Interactions with `ioreg`."""
import plistlib
import subprocess
from typing import Any, Dict, Iterable, List, Optional, cast
from xml.parsers.expat import ExpatError


def get_data(device_name: str) -> List[Dict]:
    """Returns parsed output of `ioreg` call for a given device name."""
    output = subprocess.check_output(["ioreg", "-a", "-r", "-n", device_name, "-l"])
    if output:
        try:
            return cast(List[Dict], plistlib.loads(output))
        except ExpatError:
            # Some devices seem to produce corrupt data
            pass
    return []


def get_io_dialin_device(device_name: str) -> Optional[str]:
    """Returns the value of "IODialinDevice" for a given device name."""
    ioreg_data = get_data(device_name)
    dialin_device: Optional[str] = _find_first_property_value("IODialinDevice", ioreg_data)
    return dialin_device


def _find_first_property_value(property_name: str, data: Iterable[Dict]) -> Any:
    """Finds a first value of a given proprety name in data from `ioreg`, returns None if not found."""
    found_value = None
    for item in data:
        found_value = item.get(
            property_name,
            _find_first_property_value(property_name=property_name, data=item.get("IORegistryEntryChildren", [])),
        )
        if found_value:
            break
    return found_value



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Configuration assembly algorithm."""
from pathlib import Path
from typing import Iterable, Optional

from mbed_tools.build._internal.config.config import Config
from mbed_tools.build._internal.config.cumulative_data import CumulativeData
from mbed_tools.build._internal.config.source import Source
from mbed_tools.build._internal.find_files import LabelFilter, filter_files, find_files


def assemble_config(target_attributes: dict, mbed_program_directory: Path, mbed_app_file: Optional[Path]) -> Config:
    """Assemble Config for given target and program directory.

    The structure and configuration of MbedOS requires us to do multiple passes over
    configuration files, as each pass might affect which configuration files should be included
    in the final configuration.
    """
    mbed_lib_files = find_files("mbed_lib.json", mbed_program_directory)
    return _assemble_config_from_sources_and_lib_files(target_attributes, mbed_lib_files, mbed_app_file)


def _assemble_config_from_sources_and_lib_files(
    target_attributes: dict, mbed_lib_files: Iterable[Path], mbed_app_file: Optional[Path] = None
) -> Config:
    previous_cumulative_data = None
    target_source = Source.from_target(target_attributes)
    current_cumulative_data = CumulativeData.from_sources([target_source])
    while previous_cumulative_data != current_cumulative_data:
        current_labels = current_cumulative_data.labels | current_cumulative_data.extra_labels
        filtered_files = _filter_files(
            mbed_lib_files, current_labels, current_cumulative_data.features, current_cumulative_data.components
        )
        mbed_lib_sources = [Source.from_mbed_lib(file, current_labels) for file in filtered_files]
        all_sources = [target_source] + mbed_lib_sources
        if mbed_app_file:
            all_sources = all_sources + [Source.from_mbed_app(mbed_app_file, current_labels)]

        previous_cumulative_data = current_cumulative_data
        current_cumulative_data = CumulativeData.from_sources(all_sources)

    _update_target_attributes(target_attributes, current_cumulative_data)

    return Config.from_sources(all_sources)


def _filter_files(
    files: Iterable[Path], labels: Iterable[str], features: Iterable[str], components: Iterable[str]
) -> Iterable[Path]:
    filters = (
        LabelFilter("TARGET", labels),
        LabelFilter("FEATURE", features),
        LabelFilter("COMPONENT", components),
    )
    return filter_files(files, filters)


def _update_target_attributes(target_attributes: dict, cumulative_data: CumulativeData) -> None:
    """Update target attributes with data gathered from the config system."""
    target_attributes["labels"] = cumulative_data.labels
    target_attributes["extra_labels"] = cumulative_data.extra_labels
    target_attributes["features"] = cumulative_data.features
    target_attributes["components"] = cumulative_data.components
    target_attributes["macros"] = cumulative_data.macros



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Exceptions raised by mbed tools."""


class ToolsError(Exception):
    """Base class for tools errors."""



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Command to list all Mbed enabled devices connected to the host computer."""
import click
import json
from operator import attrgetter
from typing import Iterable
from tabulate import tabulate

from mbed_tools.devices import get_connected_devices, Device
from mbed_tools.targets import Board


@click.command()
@click.option(
    "--format", type=click.Choice(["table", "json"]), default="table", show_default=True, help="Set output format."
)
@click.option(
    "--show-all",
    "-a",
    is_flag=True,
    default=False,
    help="Show all connected devices, even those which are not Mbed Boards.",
)
def list_connected_devices(format: str, show_all: bool) -> None:
    """Prints connected devices."""
    connected_devices = get_connected_devices()

    if show_all:
        devices = _sort_devices(connected_devices.identified_devices + connected_devices.unidentified_devices)
    else:
        devices = _sort_devices(connected_devices.identified_devices)

    output_builders = {
        "table": _build_tabular_output,
        "json": _build_json_output,
    }
    if devices:
        output = output_builders[format](devices)
        click.echo(output)
    else:
        click.echo("No connected Mbed devices found.")


def _sort_devices(devices: Iterable[Device]) -> Iterable[Device]:
    """Sort devices by board name and then serial number (in case there are multiple boards with the same name)."""
    return sorted(devices, key=attrgetter("mbed_board.board_name", "serial_number"))


def _build_tabular_output(devices: Iterable[Device]) -> str:
    headers = ["Board name", "Serial number", "Serial port", "Mount point(s)", "Build target(s)"]
    devices_data = []
    for device in devices:
        devices_data.append(
            [
                device.mbed_board.board_name or "<unknown>",
                device.serial_number,
                device.serial_port or "<unknown>",
                "\n".join(str(mount_point) for mount_point in device.mount_points),
                "\n".join(_get_build_targets(device.mbed_board)),
            ]
        )
    return tabulate(devices_data, headers=headers)


def _build_json_output(devices: Iterable[Device]) -> str:
    devices_data = []
    for device in devices:
        board = device.mbed_board
        devices_data.append(
            {
                "serial_number": device.serial_number,
                "serial_port": device.serial_port,
                "mount_points": [str(m) for m in device.mount_points],
                "mbed_board": {
                    "product_code": board.product_code,
                    "board_type": board.board_type,
                    "board_name": board.board_name,
                    "mbed_os_support": board.mbed_os_support,
                    "mbed_enabled": board.mbed_enabled,
                    "build_targets": _get_build_targets(board),
                },
            }
        )
    return json.dumps(devices_data, indent=4)


def _get_build_targets(board: Board) -> Iterable[str]:
    return [f"{board.board_type}_{variant}" for variant in board.build_variant] + [board.board_type]



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""API to detect any Mbed OS devices connected to the host computer.

It is expected that this package will be used by developers of Mbed OS tooling rather than by users of Mbed OS.
This package uses the https://github.com/ARMmbed/mbed-targets interface to identify valid Mbed Enabled Devices.
Please see the documentation for mbed-targets for information on configuration options.

For the command line interface to the API see the package https://github.com/ARMmbed/mbed-tools
"""
from mbed_tools.devices.devices import get_connected_devices
from mbed_tools.devices.device import Device
from mbed_tools.devices import exceptions



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Data model definition for Device and ConnectedDevices."""
from dataclasses import dataclass, field
from pathlib import Path
from typing import Tuple, Optional, List
from mbed_tools.targets import Board
from mbed_tools.devices._internal.detect_candidate_devices import CandidateDevice


@dataclass(frozen=True, order=True)
class Device:
    """Definition of an Mbed Enabled Device.

    An Mbed Device is always a USB mass storage device, which sometimes also presents a USB serial port.
    A valid Mbed Device must have a Board associated with it.

    Attributes:
        mbed_board: The Board associated with this device.
        serial_number: The serial number presented by the device to the USB subsystem.
        serial_port: The serial port presented by this device, could be None.
        mount_points: The filesystem mount points associated with this device.
    """

    mbed_board: Board
    serial_number: str
    serial_port: Optional[str]
    mount_points: Tuple[Path, ...]


@dataclass(order=True)
class ConnectedDevices:
    """Definition of connected devices which may be Mbed Boards.

    If a connected device is identified as an Mbed Board by using the HTM file on the USB mass storage device (or
    sometimes by using the serial number), it will be included in the `identified_devices` list.

    However, if the device appears as if it could be an Mbed Board but it has not been possible to find a matching
    entry in the database then it will be included in the `unidentified_devices` list.

    Attributes:
        identified_devices: A list of devices that have been identified as MbedTargets.
        unidentified_devices: A list of devices that could potentially be MbedTargets.
    """

    identified_devices: List[Device] = field(default_factory=list)
    unidentified_devices: List[Device] = field(default_factory=list)

    def add_device(self, candidate_device: CandidateDevice, mbed_board: Optional[Board] = None) -> None:
        """Add a candidate device and optionally an Mbed Target to the connected devices.

        Args:
            candidate_device: a CandidateDevice object containing the device information.
            mbed_board: a Board object for identified devices, for unidentified devices this will be None.
        """
        new_device = Device(
            serial_port=candidate_device.serial_port,
            serial_number=candidate_device.serial_number,
            mount_points=candidate_device.mount_points,
            # Create an empty Board to ensure the device is fully populated and rendering is simple
            mbed_board=mbed_board if mbed_board is not None else Board.from_offline_board_entry({}),
        )

        if mbed_board is None:
            # Keep a list of devices that could not be identified but are Mbed Boards
            self.unidentified_devices.append(new_device)
        else:
            # Keep a list of devices that have been identified as Mbed Boards
            self.identified_devices.append(new_device)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import json
import pathlib
import tempfile
from unittest import TestCase

from mbed_tools.build._internal.config.source import (
    Source,
    _namespace_data,
    _filter_target_overrides,
    _decode_json_file,
)


class TestSource(TestCase):
    def test_from_mbed_lib(self):
        with tempfile.TemporaryDirectory() as directory:
            data = {
                "name": "foo",
                "config": {"a-number": 123, "a-bool": {"help": "Simply a boolean", "value": True}},
                "target_overrides": {
                    "*": {"a-number": 456, "target.features_add": ["FOO"]},
                    "NOT_THIS_TARGET": {"a-string": "foo", "target.features_add": ["BAR"]},
                    "THIS_TARGET": {"a-bool": False, "other-lib.something-else": "blah"},
                },
                "macros": ["MACRO=1"],
            }
            file = pathlib.Path(directory, "mbed_lib.json")
            file.write_text(json.dumps(data))

            subject = Source.from_mbed_lib(file, ["THIS_TARGET"])

        self.assertEqual(
            subject,
            Source(
                human_name=f"File: {file}",
                config={"foo.a-number": 123, "foo.a-bool": {"help": "Simply a boolean", "value": True}},
                overrides={
                    "foo.a-number": 456,
                    "foo.a-bool": False,
                    "other-lib.something-else": "blah",
                    "target.features_add": ["FOO"],
                },
                macros=["MACRO=1"],
            ),
        )

    def test_from_mbed_app(self):
        with tempfile.TemporaryDirectory() as directory:
            data = {
                "config": {"a-bool": False, "a-number": {"help": "Simply a number", "value": 0}},
                "target_overrides": {
                    "*": {"a-bool": True, "target.features_add": ["HAT"]},
                    "NOT_THIS_TARGET": {"a-number": 999, "target.features_add": ["BOAT"]},
                    "THIS_TARGET": {"a-number": 2, "some-lib.something-else": "blah"},
                },
                "macros": ["SOME_MACRO=2"],
            }
            file = pathlib.Path(directory, "mbed_app.json")
            file.write_text(json.dumps(data))

            subject = Source.from_mbed_app(file, ["THIS_TARGET"])

        self.assertEqual(
            subject,
            Source(
                human_name=f"File: {file}",
                config={"app.a-bool": False, "app.a-number": {"help": "Simply a number", "value": 0}},
                overrides={
                    "app.a-number": 2,
                    "app.a-bool": True,
                    "some-lib.something-else": "blah",
                    "target.features_add": ["HAT"],
                },
                macros=["SOME_MACRO=2"],
            ),
        )

    def test_from_target(self):
        # Warning: Target is a dataclass and dataclasses provide no type safety when mocking
        target = dict(
            features={"feature_1"},
            components={"component_1"},
            labels={"label_1"},
            extra_labels={"label_2"},
            config={"foo": "bar", "target.bool": True},
            macros=["MACRO_A"],
        )

        subject = Source.from_target(target)

        self.assertEqual(
            subject,
            Source(
                human_name=f"mbed_target.Target for {target}",
                config={"target.foo": "bar", "target.bool": True},
                overrides={
                    "target.features": target["features"],
                    "target.components": target["components"],
                    "target.labels": target["labels"],
                    "target.extra_labels": target["extra_labels"],
                    "target.macros": target["macros"],
                },
                macros=[],
            ),
        )


class TestFilterTargetOverrides(TestCase):
    def test_returns_overrides_only_for_given_labels(self):
        subject = _filter_target_overrides(
            {"*": {"number": 123}, "B_TARGET": {"string": "boat"}, "A_TARGET": {"bool": True}}, ["A_TARGET"]
        )

        self.assertEqual(subject, {"number": 123, "bool": True})


class TestNamespaceData(TestCase):
    def test_prefixes_keys_without_namespace(self):
        data = {
            "foo": True,
            "hat.bar": 123,
        }

        self.assertEqual(_namespace_data(data, "my-prefix"), {"my-prefix.foo": True, "hat.bar": 123})


class TestDecodeJSONFile(TestCase):
    def test_logs_path_and_raises_when_decode_fails(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            text = "This is not valid JSON<>>"
            tmp_file = pathlib.Path(tmp_dir, "mbed_lib.json")
            tmp_file.write_text(text)

            with self.assertRaises(json.JSONDecodeError):
                with self.assertLogs(level="ERROR") as logger:
                    _decode_json_file(tmp_file)

                    self.assertIn(str(tmp_file), logger.output)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a device detector for Windows."""
from pathlib import Path
from typing import List

from mbed_tools.devices._internal.base_detector import DeviceDetector
from mbed_tools.devices._internal.candidate_device import CandidateDevice
from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader
from mbed_tools.devices._internal.windows.usb_data_aggregation import SystemUsbData, AggregatedUsbData


class WindowsDeviceDetector(DeviceDetector):
    """Windows specific implementation of device detection."""

    def __init__(self) -> None:
        """Initialiser."""
        self._data_loader = SystemDataLoader()

    def find_candidates(self) -> List[CandidateDevice]:
        """Return a generator of Candidates."""
        return [
            WindowsDeviceDetector.map_to_candidate(usb)
            for usb in SystemUsbData(data_loader=self._data_loader).all()
            if WindowsDeviceDetector.is_valid_candidate(usb)
        ]

    @staticmethod
    def map_to_candidate(usb_data: AggregatedUsbData) -> CandidateDevice:
        """Maps a USB device to a candidate."""
        serial_port = next(iter(usb_data.get("serial_port")), None)
        uid = usb_data.uid
        return CandidateDevice(
            product_id=uid.product_id,
            vendor_id=uid.vendor_id,
            mount_points=tuple(Path(disk.component_id) for disk in usb_data.get("disks")),
            serial_number=uid.uid.presumed_serial_number,
            serial_port=serial_port.port_name if serial_port else None,
        )

    @staticmethod
    def is_valid_candidate(usb_data: AggregatedUsbData) -> bool:
        """States whether the usb device is a valid candidate or not."""
        return usb_data.is_composite and usb_data.is_associated_with_disk



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import factory

from mbed_tools.build._internal.config.config import Config, Option, Macro
from mbed_tools.build._internal.config.source import Source


class SourceFactory(factory.Factory):
    class Meta:
        model = Source

    human_name = factory.Faker("name")
    config = factory.Dict({})
    overrides = factory.Dict({})
    macros = factory.List([])


class OptionFactory(factory.Factory):
    class Meta:
        model = Option

    value = "0"
    macro_name = "MACRO_NAME"
    help_text = factory.Faker("text")
    set_by = "libname"
    key = "key"


class MacroFactory(factory.Factory):
    class Meta:
        model = Macro

    value = "0"
    name = "A_MACRO"
    set_by = "source"


class ConfigFactory(factory.Factory):
    class Meta:
        model = Config

    options = factory.Dict({})
    macros = factory.Dict({})



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Build configuration representation."""
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, Optional

from mbed_tools.build._internal.config.source import Source
from mbed_tools.build._internal.config.cumulative_data import CUMULATIVE_OVERRIDE_KEYS_IN_SOURCE
from mbed_tools.build._internal.config.bootloader_overrides import BOOTLOADER_OVERRIDE_KEYS_IN_SOURCE


@dataclass
class Macro:
    """Representation of a macro."""

    value: Any
    name: str
    set_by: str

    @classmethod
    def build(cls, macro: str, source: Source) -> "Macro":
        """Build macro from macro string.

        There are two flavours of macro strings in MbedOS configuration:
        - with value: FOO=BAR
        - without value: FOO
        """
        value: Any
        if "=" in macro:
            name, value = macro.split("=")
        else:
            name = macro
            value = None

        return cls(name=name, value=value, set_by=source.human_name)


@dataclass
class Option:
    """Representation of a configuration option."""

    value: Any
    macro_name: Optional[str]
    help_text: Optional[str]
    set_by: str
    key: str

    @classmethod
    def build(cls, key: str, data: Any, source: Source) -> "Option":
        """Build configuration option from config entry value.

        Config values are either complex data structures or simple values.
        This function handles both.

        Args:
            key: Namespaced configuration key
            data: Configuration data - a dict or a primitive
            source: Source from which option data came - used for tracing overrides
        """
        if isinstance(data, dict):
            return cls(
                key=key,
                value=_sanitize_option_value(data.get("value")),
                macro_name=data.get("macro_name", _build_option_macro_name(key)),
                help_text=data.get("help"),
                set_by=source.human_name,
            )
        else:
            return cls(
                value=_sanitize_option_value(data),
                key=key,
                macro_name=_build_option_macro_name(key),
                help_text=None,
                set_by=source.human_name,
            )

    def set_value(self, value: Any, source: Source) -> "Option":
        """Mutate self with new value."""
        self.value = _sanitize_option_value(value)
        self.set_by = source.human_name
        return self


IGNORED_OVERRIDE_KEYS_IN_SOURCE = CUMULATIVE_OVERRIDE_KEYS_IN_SOURCE + BOOTLOADER_OVERRIDE_KEYS_IN_SOURCE


@dataclass
class Config:
    """Representation of config attributes assembled during Source parsing.

    Attributes:
        options: Options parsed from "config" and "target_overrides" sections of *.json files
        macros: Macros parsed from "macros" section of mbed_lib.json and mbed_app.json file
    """

    options: Dict[str, Option] = field(default_factory=dict)
    macros: Dict[str, Macro] = field(default_factory=dict)

    @classmethod
    def from_sources(cls, sources: Iterable[Source]) -> "Config":
        """Interrogate each source in turn to create final Config."""
        config = Config()
        for source in sources:
            for key, value in source.config.items():
                # If the config item is about a certain component or feature
                # being present, avoid adding it to the mbed_config.cmake
                # configuration file. Instead, applications should depend on
                # the feature or component with target_link_libraries() and the
                # component's CMake flle (in the Mbed OS repo) will create
                # any necessary macros or definitions.
                if key.endswith(".present"):
                    continue
                _create_config_option(config, key, value, source)
            for key, value in source.overrides.items():
                if key in IGNORED_OVERRIDE_KEYS_IN_SOURCE:
                    continue
                _update_config_option(config, key, value, source)
            for value in source.macros:
                _create_macro(config, value, source)
        return config


def _create_config_option(config: Config, key: str, value: Any, source: Source) -> None:
    """Mutates Config in place by creating a new Option."""
    config.options[key] = Option.build(key, value, source)


def _update_config_option(config: Config, key: str, value: Any, source: Source) -> None:
    """Mutates Config in place by updating the value of existing Option."""
    if key not in config.options:
        raise ValueError(
            f"Can't update option which does not exist."
            f" Attempting to set '{key}' to '{value}' in '{source.human_name}'."
        )
    config.options[key].set_value(value, source)


def _build_option_macro_name(config_key: str) -> str:
    """Build macro name for configuration key.

    All configuration variables require a macro name, so that they can be referenced in a header file.
    Some values in config define "macro_name", some don't. This helps generate consistent macro names
    for the latter.
    """
    sanitised_config_key = config_key.replace(".", "_").replace("-", "_").upper()
    return f"MBED_CONF_{sanitised_config_key}"


def _sanitize_option_value(value: Any) -> Any:
    """Converts booleans to ints, leaves everything else as is."""
    if isinstance(value, bool):
        return int(value)
    else:
        return value


def _create_macro(config: Config, macro_str: str, source: Source) -> None:
    """Mutates Config in place by creating a new macro."""
    macro = Macro.build(macro_str, source)
    existing = config.macros.get(macro.name)
    if existing:
        raise ValueError(
            f"Can't override previously set macro."
            f" Attempting to set '{macro.name}' to '{macro.value}' in '{source.human_name}'."
            f" Set previously by '{existing.set_by}'."
        )
    config.macros[macro.name] = macro



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase, mock

import click
from click.testing import CliRunner

from mbed_tools.cli import cli
from mbed_tools.cli.list_connected_devices import list_connected_devices
from mbed_tools.lib.exceptions import ToolsError


class TestDevicesCommandIntegration(TestCase):
    def test_devices_is_integrated(self):
        self.assertEqual(cli.commands["devices"], list_connected_devices)


class TestClickGroupWithExceptionHandling(TestCase):
    @mock.patch("mbed_tools.cli.LOGGER.error", autospec=True)
    def test_logs_tools_errors(self, logger_error):
        def callback():
            raise ToolsError()

        mock_cli = click.Command("test", callback=callback)
        cli.add_command(mock_cli, "test")

        runner = CliRunner()
        result = runner.invoke(cli, ["test"])

        self.assertEqual(1, result.exit_code)
        logger_error.assert_called_once()


class TestVersionCommand(TestCase):
    def test_version_command(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["--version"])
        self.assertTrue(result.output)
        self.assertEqual(0, result.exit_code)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from mbed_tools.targets import Board


def make_board(
    board_type="BoardType",
    board_name="BoardName",
    mbed_os_support=None,
    mbed_enabled=None,
    product_code="9999",
    slug="BoardSlug",
    target_type="TargetType",
):
    return Board(
        board_type=board_type,
        product_code=product_code,
        board_name=board_name,
        target_type=target_type,
        slug=slug,
        mbed_os_support=mbed_os_support if mbed_os_support else (),
        mbed_enabled=mbed_enabled if mbed_enabled else (),
        build_variant=(),
    )


def make_dummy_internal_board_data():
    return [dict(attributes=dict(board_type=str(i), board_name=str(i), product_code=str(i))) for i in range(10)]



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tag a new version based on the presence of news fragments.

mbed-tools follows a semantic versioning scheme based on the extension of the news file added to a commit.

The following versioning scheme is used:

    Extension  | Version Bumped
    ---------------------------
     .major    | Major version
     .feature  | Minor version
     .bugfix   | Patch version
     .doc      | N/A
     .removal  | N/A
     .misc     | N/A

This script will look at the latest tag reachable from the current commit using `git describe`, it will then look in
the news fragments directory, passed on the command line, and deduce which part of the version needs to be bumped based
on the news file extension.
"""
import argparse
import logging
import pathlib
import sys

import git
import semver

from typing import Iterable, Tuple


def bump_version_from_news_files(current_version: str, news_files: Iterable[pathlib.Path]) -> str:
    """Bumps the current version based on the given list of news files.

    Looks at the list of news files, determines which part of the version to bump based on the extension.

    Args:
        current_version: The current version. Does not accept all pep440 version strings.
        news_files: Iterable of news files used to work out which part of the current_version to bump.

    Returns:
        The new version as a semver formatted string.
    """
    semver_to_bump = news_ext_to_version_type(tuple(news.suffix for news in news_files if news.suffix != ".md"))
    version_info = semver.VersionInfo.parse(current_version)
    if semver_to_bump == "major":
        logging.info("Bumping major version.")
        version_info = version_info.bump_major()
    elif semver_to_bump == "minor":
        logging.info("Bumping minor version.")
        version_info = version_info.bump_minor()
    elif semver_to_bump == "patch":
        logging.info("Bumping patch version.")
        version_info = version_info.bump_patch()

    new_version = f"{version_info.major}.{version_info.minor}.{version_info.patch}"
    logging.info("New version is: %s", new_version)
    return new_version


def news_ext_to_version_type(news_file_exts: Tuple[str]) -> str:
    """Given a tuple of news file extensions, work out which part of the version to bump.

    Args:
        news_file_exts: The news file extensions to be considered.

    Returns:
        The part of the version to bump as a string.
    """
    logging.debug("News file extensions found: %s", set(news_file_exts))
    if ".major" in news_file_exts:
        return "major"
    elif ".feature" in news_file_exts:
        return "minor"
    elif ".bugfix" in news_file_exts:
        return "patch"
    raise ValueError("No significant news file extensions found. Version will not be incremented.")


def parse_args() -> argparse.Namespace:
    """Define the command line argument parser.

    Returns:
        argparse.Namespace containing references to the values of known command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n", "--news-file-directory", type=pathlib.Path, required=True, help="Directory where news files are located."
    )
    parser.add_argument("-v", "--verbosity", action="count", default=0, help="Set verbosity level.")
    return parser.parse_args()


def set_log_level(verbosity: int) -> None:
    """Set the log level based on the verbosity.

    The log level is set as follows:
    verbosity = 0 -> logging.ERROR
    verbosity = 1 -> logging.WARN
    verbosity = 2 -> logging.INFO
    verbosity = 3 -> logging.DEBUG
    """
    levels = (logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG)
    logging.getLogger().setLevel(levels[min(verbosity, len(levels) - 1)])


def get_current_version():
    """Get the current version using git describe to check the latest tag."""
    with git.Repo() as repo:
        return repo.git.describe("--tags")


def main(args: argparse.Namespace) -> int:
    """Entry point.

    Args:
        args: Parsed command line arguments.
    """
    try:
        set_log_level(args.verbosity)
        current_version = get_current_version()
        logging.info("Current version is %s", current_version)
        new_version = bump_version_from_news_files(current_version, list(args.news_file_directory.iterdir()))
        print(new_version)
        return 0
    except Exception as e:
        logging.error(e)
        return 1


if __name__ == "__main__":
    sys.exit(main(parse_args()))



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase, mock

from mbed_tools.build import generate_config


class TestGenerateConfig(TestCase):
    @mock.patch("mbed_tools.build.config.MbedProgram")
    @mock.patch("mbed_tools.build.config.get_target_by_name")
    @mock.patch("mbed_tools.build.config.generate_mbed_config_cmake_file")
    @mock.patch("mbed_tools.build.config.assemble_config")
    @mock.patch("mbed_tools.build.config.write_file")
    def test_collaborators_called_with_corrrect_arguments(
        self, write_file, assemble_config, gen_config_cmake, get_target_by_name, mbed_program
    ):
        program = mbed_program.from_existing()

        generate_config("K64F", "GCC_ARM", program)

        get_target_by_name.assert_called_once_with("K64F", program.mbed_os.targets_json_file)
        assemble_config.assert_called_once_with(
            get_target_by_name.return_value, program.root, program.files.app_config_file,
        )
        gen_config_cmake.assert_called_once_with(
            "K64F", get_target_by_name.return_value, assemble_config.return_value, "GCC_ARM"
        )
        write_file.assert_called_once_with(
            program.files.cmake_config_file.parent, program.files.cmake_config_file.name, gen_config_cmake.return_value,
        )



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Interface for device detectors."""
from abc import ABC, abstractmethod
from typing import List

from mbed_tools.devices._internal.candidate_device import CandidateDevice


class DeviceDetector(ABC):
    """Object in charge of finding USB devices."""

    @abstractmethod
    def find_candidates(self) -> List[CandidateDevice]:
        """Returns CandidateDevices."""
        pass



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Interface for accessing Targets from Mbed OS's targets.json.

An instance of `mbed_tools.targets.target.Target`
can be retrieved by calling one of the public functions.
"""
import pathlib

from mbed_tools.targets.exceptions import TargetError
from mbed_tools.targets._internal import target_attributes


def get_target_by_name(name: str, path_to_targets_json: pathlib.Path) -> dict:
    """Returns a dictionary of attributes for the target whose name matches the name given.

    The target is as defined in the targets.json file found in the Mbed OS library.

    Args:
        name: the name of the Target to be returned
        path_to_targets_json: path to a targets.json file containing target definitions

    Raises:
        TargetError: an error has occurred while fetching target
    """
    try:
        return target_attributes.get_target_attributes(path_to_targets_json, name)
    except (FileNotFoundError, target_attributes.TargetAttributesError) as e:
        raise TargetError(e) from e


def get_target_by_board_type(board_type: str, path_to_targets_json: pathlib.Path) -> dict:
    """Returns the target whose name matches a board's build_type.

    The target is as defined in the targets.json file found in the Mbed OS library.

    Args:
        board_type: a board's board_type (see `mbed_tools.targets.board.Board`)
        path_to_targets_json: path to a targets.json file containing target definitions

    Raises:
        TargetError: an error has occurred while fetching target
    """
    return get_target_by_name(board_type, path_to_targets_json)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib
import tempfile
from unittest import TestCase, mock
from mbed_tools.targets.exceptions import UnknownBoard

from tests.devices.factories import CandidateDeviceFactory
from mbed_tools.devices._internal.htm_file import OnlineId
from mbed_tools.devices._internal.resolve_board import (
    NoBoardForCandidate,
    resolve_board,
    _get_all_htm_files_contents,
    _read_htm_file_contents,
    _is_htm_file,
)


@mock.patch(
    "mbed_tools.devices._internal.resolve_board._get_all_htm_files_contents",
    autospec=True,
    return_value=["some file contents"],
)
@mock.patch("mbed_tools.devices._internal.resolve_board.read_product_code", autospec=True)
@mock.patch("mbed_tools.devices._internal.resolve_board.get_board_by_product_code", autospec=True)
class TestResolveBoardUsingProductCodeFromHTM(TestCase):
    def test_returns_resolved_target(self, get_board_by_product_code, read_product_code, _get_all_htm_files_contents):
        read_product_code.return_value = "0123"
        candidate = CandidateDeviceFactory()

        subject = resolve_board(candidate)

        self.assertEqual(subject, get_board_by_product_code.return_value)
        get_board_by_product_code.assert_called_once_with(read_product_code.return_value)
        read_product_code.assert_called_once_with(_get_all_htm_files_contents.return_value[0])
        _get_all_htm_files_contents.assert_called_once_with(candidate.mount_points)

    def test_raises_when_board_not_found(
        self, get_board_by_product_code, read_product_code, _get_all_htm_files_contents
    ):
        read_product_code.return_value = "1234"
        get_board_by_product_code.side_effect = UnknownBoard
        candidate = CandidateDeviceFactory()

        with self.assertRaises(NoBoardForCandidate):
            resolve_board(candidate)


@mock.patch(
    "mbed_tools.devices._internal.resolve_board._get_all_htm_files_contents",
    autospec=True,
    return_value=["other file contents"],
)
@mock.patch("mbed_tools.devices._internal.resolve_board.read_product_code", autospec=True, return_value=None)
@mock.patch("mbed_tools.devices._internal.resolve_board.read_online_id", autospec=True)
@mock.patch("mbed_tools.devices._internal.resolve_board.get_board_by_online_id", autospec=True)
class TestResolveBoardUsingOnlineIdFromHTM(TestCase):
    def test_returns_resolved_board(
        self, get_board_by_online_id, read_online_id, read_product_code, _get_all_htm_files_contents
    ):
        online_id = OnlineId(target_type="hat", slug="boat")
        read_online_id.return_value = online_id
        candidate = CandidateDeviceFactory()

        subject = resolve_board(candidate)

        self.assertEqual(subject, get_board_by_online_id.return_value)
        read_online_id.assert_called_with(_get_all_htm_files_contents.return_value[0])
        get_board_by_online_id.assert_called_once_with(target_type=online_id.target_type, slug=online_id.slug)

    def test_raises_when_board_not_found(
        self, get_board_by_online_id, read_online_id, read_product_code, _get_all_htm_files_contents
    ):
        read_online_id.return_value = OnlineId(target_type="hat", slug="boat")
        get_board_by_online_id.side_effect = UnknownBoard
        candidate = CandidateDeviceFactory()

        with self.assertRaises(NoBoardForCandidate):
            resolve_board(candidate)


@mock.patch(
    "mbed_tools.devices._internal.resolve_board._get_all_htm_files_contents",
    autospec=True,
    return_value=["who knows file contents"],
)
@mock.patch("mbed_tools.devices._internal.resolve_board.read_product_code", autospec=True, return_value=None)
@mock.patch("mbed_tools.devices._internal.resolve_board.read_online_id", autospec=True, return_value=None)
@mock.patch("mbed_tools.devices._internal.resolve_board.get_board_by_product_code", autospec=True)
class TestResolveBoardUsingProductCodeFromSerial(TestCase):
    def test_resolves_board_using_product_code_when_available(
        self, get_board_by_product_code, read_online_id, read_product_code, _get_all_htm_files_contents
    ):
        candidate = CandidateDeviceFactory()

        subject = resolve_board(candidate)

        self.assertEqual(subject, get_board_by_product_code.return_value)
        get_board_by_product_code.assert_called_once_with(candidate.serial_number[:4])

    def test_raises_when_board_not_found(
        self, get_board_by_product_code, read_online_id, read_product_code, _get_all_htm_files_contents
    ):
        get_board_by_product_code.side_effect = UnknownBoard
        candidate = CandidateDeviceFactory()

        with self.assertRaises(NoBoardForCandidate):
            resolve_board(candidate)


class TestGetAllHtmFilesContents(TestCase):
    def test_returns_contents_of_all_htm_files_in_given_directories(self):
        with tempfile.TemporaryDirectory() as directory:
            directory_1 = pathlib.Path(directory, "test-1")
            directory_1.mkdir()
            directory_2 = pathlib.Path(directory, "test-2")
            directory_2.mkdir()
            pathlib.Path(directory_1, "mbed.htm").write_text("foo")
            pathlib.Path(directory_2, "whatever.htm").write_text("bar")
            pathlib.Path(directory_1, "file.txt").write_text("txt files should not be read")
            pathlib.Path(directory_1, "._MBED.HTM").write_text("hidden files should not be read")

            result = _get_all_htm_files_contents([directory_1, directory_2])

        self.assertEqual(result, ["foo", "bar"])


class TestReadHtmFilesContents(TestCase):
    def test_handles_unreadable_htm_file(self):
        with tempfile.TemporaryDirectory() as directory:
            htm_file = pathlib.Path(directory, "mbed.htm")
            htm_file.write_text("foo")

            result = _read_htm_file_contents([htm_file, pathlib.Path("error.htm")])

        self.assertEqual(result, ["foo"])


class TestIsHtmFile(TestCase):
    def test_lower_case_htm(self):
        result = _is_htm_file(pathlib.Path("mbed.htm"))
        self.assertEqual(True, result)

    def test_upper_case_htm(self):
        result = _is_htm_file(pathlib.Path("MBED.HTM"))
        self.assertEqual(True, result)

    def test_hidden_htm(self):
        result = _is_htm_file(pathlib.Path(".htm"))
        self.assertEqual(False, result)

    def test_text_file(self):
        result = _is_htm_file(pathlib.Path("mbed.txt"))
        self.assertEqual(False, result)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Interface to the Board Database."""
import json

from dataclasses import asdict
from collections.abc import Set
from typing import Iterator, Iterable, Any, Callable

from mbed_tools.targets._internal import board_database

from mbed_tools.targets.exceptions import UnknownBoard
from mbed_tools.targets.board import Board


class Boards(Set):
    """Interface to the Board Database.

    Boards is initialised with an Iterable[Board]. The classmethods
    can be used to construct Boards with data from either the online or offline database.
    """

    @classmethod
    def from_offline_database(cls) -> "Boards":
        """Initialise with the offline board database.

        Raises:
            BoardDatabaseError: Could not retrieve data from the board database.
        """
        return cls(Board.from_offline_board_entry(b) for b in board_database.get_offline_board_data())

    @classmethod
    def from_online_database(cls) -> "Boards":
        """Initialise with the online board database.

        Raises:
            BoardDatabaseError: Could not retrieve data from the board database.
        """
        return cls(Board.from_online_board_entry(b) for b in board_database.get_online_board_data())

    def __init__(self, boards_data: Iterable["Board"]) -> None:
        """Initialise with a list of boards.

        Args:
            boards_data: iterable of board data from a board database source.
        """
        self._boards_data = tuple(boards_data)

    def __iter__(self) -> Iterator["Board"]:
        """Yield an Board on each iteration."""
        for board in self._boards_data:
            yield board

    def __len__(self) -> int:
        """Return the number of boards."""
        return len(self._boards_data)

    def __contains__(self, board: object) -> Any:
        """Check if a board is in the collection of boards.

        Args:
            board: An instance of Board.
        """
        if not isinstance(board, Board):
            return False

        return any(x == board for x in self)

    def get_board(self, matching: Callable) -> Board:
        """Returns first Board for which `matching` returns True.

        Args:
            matching: A function which will be called for each board in database

        Raises:
            UnknownBoard: the given product code was not found in the board database.
        """
        try:
            return next(board for board in self if matching(board))
        except StopIteration:
            raise UnknownBoard()

    def json_dump(self) -> str:
        """Return the contents of the board database as a json string."""
        return json.dumps([asdict(b) for b in self], indent=4)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase, mock

from mbed_tools.targets import Board
from mbed_tools.targets.exceptions import MbedTargetsError

from tests.devices.factories import CandidateDeviceFactory
from mbed_tools.devices.device import Device
from mbed_tools.devices._internal.exceptions import NoBoardForCandidate

from mbed_tools.devices.devices import get_connected_devices
from mbed_tools.devices.exceptions import DeviceLookupFailed


@mock.patch("mbed_tools.devices.devices.detect_candidate_devices")
@mock.patch("mbed_tools.devices.devices.resolve_board")
class TestGetConnectedDevices(TestCase):
    def test_builds_devices_from_candidates(self, resolve_board, detect_candidate_devices):
        candidate = CandidateDeviceFactory()
        detect_candidate_devices.return_value = [candidate]

        connected_devices = get_connected_devices()
        self.assertEqual(
            connected_devices.identified_devices,
            [
                Device(
                    serial_port=candidate.serial_port,
                    serial_number=candidate.serial_number,
                    mount_points=candidate.mount_points,
                    mbed_board=resolve_board.return_value,
                )
            ],
        )
        self.assertEqual(connected_devices.unidentified_devices, [])
        resolve_board.assert_called_once_with(candidate)

    @mock.patch.object(Board, "from_offline_board_entry")
    def test_skips_candidates_without_a_board(self, board, resolve_board, detect_candidate_devices):
        candidate = CandidateDeviceFactory()
        resolve_board.side_effect = NoBoardForCandidate
        detect_candidate_devices.return_value = [candidate]
        board.return_value = None

        connected_devices = get_connected_devices()
        self.assertEqual(connected_devices.identified_devices, [])
        self.assertEqual(
            connected_devices.unidentified_devices,
            [
                Device(
                    serial_port=candidate.serial_port,
                    serial_number=candidate.serial_number,
                    mount_points=candidate.mount_points,
                    mbed_board=None,
                )
            ],
        )

    def test_raises_device_lookup_failed_on_internal_error(self, resolve_board, detect_candidate_devices):
        resolve_board.side_effect = MbedTargetsError
        detect_candidate_devices.return_value = [CandidateDeviceFactory()]

        with self.assertRaises(DeviceLookupFailed):
            get_connected_devices()



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Functions for parsing the inheritance for overriding attributes.

Overriding attributes are defined and can be overridden further down the hierarchy.

There are two types - standard and merging. If an attribute is defined as a
merging attribute, only individual attribute elements override, otherwise
the entire attribute is overridden by any later definitions.

The hierarchy is determined as 'depth-first' in multiple inheritance.
eg. Look first for the attribute in the current target. If not found,
look for the attribute in the first target's parent, then in the parent of the parent and so on.
If not found, look for the property in the rest of the target's parents, relative to the
current inheritance level.

This means a target on a higher level could potentially override one on a lower level.
"""
from collections import deque
from functools import reduce
from typing import Dict, List, Any, Deque, Set

from mbed_tools.targets._internal.targets_json_parsers.accumulating_attribute_parser import ALL_ACCUMULATING_ATTRIBUTES

MERGING_ATTRIBUTES = ("config", "overrides")
NON_OVERRIDING_ATTRIBUTES = ALL_ACCUMULATING_ATTRIBUTES + ("public", "inherits")


def get_overriding_attributes_for_target(all_targets_data: Dict[str, Any], target_name: str) -> Dict[str, Any]:
    """Parses the data for all targets and returns the overriding attributes for the specified target.

    Args:
        all_targets_data: a dictionary representation of the contents of targets.json
        target_name: the name of the target to find the attributes of

    Returns:
        A dictionary containing all the overriding attributes for the chosen target
    """
    override_order = _targets_override_hierarchy(all_targets_data, target_name)
    return _determine_overridden_attributes(override_order)


def get_labels_for_target(all_targets_data: Dict[str, Any], target_name: str) -> Set[str]:
    """The labels for a target are the names of all the boards (public and private) that the board inherits from.

    The order of these labels are not reflective of inheritance order.

    Args:
        all_targets_data: a dictionary representation of the contents of targets.json
        target_name: the name of the target to find the labels for

    Returns:
        A set of names of boards that make up the inheritance tree for the target
    """
    targets_in_order = _targets_override_hierarchy(all_targets_data, target_name)
    return _extract_target_labels(targets_in_order, target_name)


def _targets_override_hierarchy(all_targets_data: Dict[str, Any], target_name: str) -> List[dict]:
    """List all ancestors of a target in order of overriding inheritance (depth-first).

    Using a depth-first traverse of the inheritance tree, return a list of targets in the
    order of inheritance, starting with the target itself and finishing with its highest ancestor.

    Eg. An inheritance tree diagram for target "A" below

    D     E
    |     |
    B     C
    |_____|
       |
       A

    Would give us an inheritance order of [A, B, D, C, E]

    Args:
        all_targets_data: a dictionary representation of all the data in a targets.json file
        target_name: the name of the target we want to calculate the attributes for

    Returns:
        A list of dicts representing each target in the hierarchy.
    """
    targets_in_order: List[dict] = []

    still_to_visit: Deque[dict] = deque()
    still_to_visit.appendleft(all_targets_data[target_name])

    while still_to_visit:
        current_target = still_to_visit.popleft()
        targets_in_order.append(current_target)
        for parent_target in reversed(current_target.get("inherits", [])):
            still_to_visit.appendleft(all_targets_data[parent_target])

    return targets_in_order


def _determine_overridden_attributes(targets_in_order: List[dict]) -> Dict[str, Any]:
    """Finds all the overrideable attributes for a target from its list of ancestors.

    Combines the attributes from all the targets in the hierarchy. Starts from the highest ancestor
    reduces down to the target itself, overriding if they define the same attribute.

    Identifies the attributes that should be merged (only their elements overridden,
    not the entire attribute definition) and updates their contents.

    Removes any accumulating attributes - they will be handled by a separate parser.

    Args:
        targets_in_order: list of targets in order of inheritance, starting with the target up to its highest ancestor

    Returns:
        A dictionary containing all the overridden attributes for a target
    """
    target_attributes = _reduce_right_list_of_dictionaries(targets_in_order)
    for merging_attribute in MERGING_ATTRIBUTES:
        override_order_for_single_attribute = [target.get(merging_attribute, {}) for target in targets_in_order]
        merged_attribute_elements = _reduce_right_list_of_dictionaries(list(override_order_for_single_attribute))
        if merged_attribute_elements:
            target_attributes[merging_attribute] = merged_attribute_elements
    overridden_attributes = _remove_unwanted_attributes(target_attributes)
    return overridden_attributes


def _reduce_right_list_of_dictionaries(list_of_dicts: List[dict]) -> Dict[str, Any]:
    """Starting from rightmost dict, merge dicts together, left dict overriding the right."""
    return reduce(lambda x, y: {**x, **y}, reversed(list_of_dicts))


def _remove_unwanted_attributes(target_attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Removes all non-overriding attributes.

    Defined in NON_OVERRIDING_ATTRIBUTES.
    Accumulating arguments are inherited in a different way that is handled by its own parser.
    The 'public' attribute is not inherited.
    The 'inherits' attribute is only needed to calculate the hierarchies.

    Args:
        target_attributes: a dictionary of attributes for a target

    Returns:
        The target attributes with the accumulating attributes removed.
    """
    output_dict = target_attributes.copy()
    for attribute in NON_OVERRIDING_ATTRIBUTES:
        output_dict.pop(attribute, None)
    return output_dict


def _extract_target_labels(targets_in_order: List[dict], target_name: str) -> Set[str]:
    """Collect a set of all the board names from the inherits field in each target in the hierarchy.

    Args:
        targets_in_order: list of targets in order of inheritance, starting with the target up to its highest ancestor
        target_name: the name of the target to find the labels for

    Returns:
        A set of names of boards that make up the inheritance tree for the target
    """
    labels = {target_name}
    for target in targets_in_order:
        for parent in target.get("inherits", []):
            labels.add(parent)
    return labels



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a USB Identifier."""

import re
from enum import Enum
from typing import Optional, Pattern, Any, List, cast

from mbed_tools.devices._internal.utils.python_helpers import named_tuple_with_defaults
from mbed_tools.devices._internal.windows.component_descriptor_utils import is_undefined_data_object
from mbed_tools.devices._internal.windows.windows_identifier import WindowsUID

KEY_UID = "UID"


class USBIdentifierToken(Enum):
    """Tokens used in the device Id field.

    See https://docs.microsoft.com/en-us/windows-hardware/drivers/install/standard-usb-identifiers
    """

    VID = 1
    PID = 2
    REV = 3
    MI = 4

    @property
    def pattern(self) -> Pattern:
        """Gets the regex pattern for the corresponding part."""
        return re.compile(f"^{self.name}_(.*)$")

    @staticmethod
    def get_patterns_dict() -> dict:
        """Returns a dictionary of all the regexes."""
        return {p: p.pattern for p in USBIdentifierToken}


class UsbIdentifier(
    named_tuple_with_defaults(  # type:ignore
        typename="UsbIdentifier",
        field_names=[KEY_UID] + [part.name for part in USBIdentifierToken],
        defaults=[None] * (len(USBIdentifierToken) + 1),
    )
):
    """Object describing the different elements present in a the device Id."""

    def get(self, key: USBIdentifierToken) -> str:
        """Returns the value corresponding to a specific token."""
        return str(getattr(self, key.name))

    @property
    def uid(self) -> WindowsUID:
        """Gets the USB ID."""
        return cast(WindowsUID, getattr(self, KEY_UID))

    def contains_genuine_serial_number(self) -> bool:
        """Contains a genuine serial number and not an instance ID."""
        return self.uid.contains_genuine_serial_number()

    @property
    def product_id(self) -> str:
        """Returns the product id field."""
        return self.get(USBIdentifierToken.PID)

    @property
    def vendor_id(self) -> str:
        """Returns the product id field."""
        return self.get(USBIdentifierToken.VID)

    def __eq__(self, other: Any) -> bool:
        """States whether the other id equals to self."""
        if not other or not isinstance(other, UsbIdentifier):
            return False
        if self.is_undefined:
            return other.is_undefined

        return all([self.uid == other.uid, self.product_id == other.product_id, self.vendor_id == other.vendor_id])

    def __hash__(self) -> int:
        """Generates a hash."""
        return hash(self.uid) + hash(self.product_id) + hash(self.vendor_id)

    @property
    def is_undefined(self) -> bool:
        """States whether none of the elements present in DeviceId were defined."""
        return is_undefined_data_object(self)


class Win32DeviceIdParser:
    """Parser of a standard Win32 device ID.

    See https://docs.microsoft.com/en-us/windows-hardware/drivers/install/standard-usb-identifiers
    """

    def parse_uid(self, raw_id: str, serial_number: Optional[str] = None) -> WindowsUID:
        """Parses the UID value.

        As described here:
        https://docs.microsoft.com/it-it/windows-hardware/drivers/install/device-instance-ids
        https://stackoverflow.com/questions/51513337/is-the-usb-instance-id-on-windows-unique-for-a-device
        https://docs.microsoft.com/it-it/windows-hardware/drivers/install/instance-ids
        the instance ID corresponds to the serial number information, if supported by the underlying bus, otherwise
        it is generated by Windows.

        For some boards (e.g. ST boards), the ID may contain other information that we are not interested in
        (e.g. MI value). This method tries to retrieve the actual ID.
        """
        id_elements = raw_id.split("&")
        if len(id_elements) <= 1:
            # The instance ID is the serial number.
            return WindowsUID(uid=raw_id.lower(), raw_uid=raw_id, serial_number=serial_number)
        # The instance ID is generated by Windows and hence might contain other element than the ParentPrefixID.
        # The following tries to only consider what may be the ParentPrefixID.
        return WindowsUID(uid="&".join(id_elements[:-1]).lower(), raw_uid=raw_id, serial_number=serial_number)

    def record_id_element(self, element: str, valuable_information: dict, patterns_dict: dict) -> None:
        """Stores recognised parts of the device ID based on patterns defined."""
        for k, p in patterns_dict.items():
            match = p.fullmatch(element)
            if match:
                valuable_information[k.name] = match.group(1)

    def split_id_elements(self, parts: List[str], serial_number: str = None) -> dict:
        """Splits the different elements of an Device ID."""
        information = dict()
        information[KEY_UID] = self.parse_uid(parts[-1], serial_number)
        other_elements = parts[-2].split("&")
        patterns_dict = USBIdentifierToken.get_patterns_dict()
        for element in other_elements:
            self.record_id_element(element, information, patterns_dict)
        return information

    def parse(self, id_string: Optional[str], serial_number: Optional[str] = None) -> "UsbIdentifier":
        r"""Parses the device id string and retrieves the different elements of interest.

        See https://docs.microsoft.com/en-us/windows-hardware/drivers/install/standard-usb-identifiers
         Format: <device-ID>\<instance-specific-ID>
         Ex. `USB\VID_2109&PID_8110\5&376ABA2D&0&21`
          - `<device-ID>`: `USB\VID_2109&PID_8110`
          - `<instance-specific-ID>`: `5&376ABA2D&0&21`
        [Device instance IDs](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/device-instance-ids)
        -> [Device IDs](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/device-ids)
        -> [Hardware IDs](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/hardware-ids)
        -> [Device identifier formats](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/device-identifier-formats)  # noqa: E501
        -> [Identifiers for USB](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/identifiers-for-usb-devices)
         - [Standard USB Identifiers](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/standard-usb-identifiers)
         - [Special USB Identifiers](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/special-usb-identifiers)
          - [Instance specific ID](https://docs.microsoft.com/en-us/windows-hardware/drivers/install/instance-ids)

        Returns:
            corresponding DeviceIdInformation.
        """
        if not id_string or len(id_string.strip()) == 0:
            return UsbIdentifier()
        parts = id_string.split("\\")
        if len(parts) < 2:
            return UsbIdentifier()
        return UsbIdentifier(**self.split_id_elements(parts, serial_number))


def parse_device_id(id_string: Optional[str], serial_number: Optional[str] = None) -> UsbIdentifier:
    """Parses the device id string and retrieves the different elements of interest.

    See https://docs.microsoft.com/en-us/windows-hardware/drivers/install/standard-usb-identifiers
    """
    return Win32DeviceIdParser().parse(id_string, serial_number)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import unittest

from mbed_tools.lib.python_helpers import minimum_python_version, named_tuple_with_defaults


class TestPythonHelpers(unittest.TestCase):
    def test_python_version(self):
        # Tools only support Python>= 3.6
        self.assertTrue(minimum_python_version(3, 6))
        # With a bit of luck, python 100.100 is not coming tomorrow
        self.assertFalse(minimum_python_version(100, 100))

    def test_named_tuple_with_defaults(self):
        named_tuple = named_tuple_with_defaults("TestNamedTuple", field_names=["field1", "field2"], defaults=[1, 2])
        a_tuple = named_tuple()
        self.assertIsNotNone(a_tuple)
        self.assertEqual(a_tuple.field1, 1)
        self.assertEqual(a_tuple.field2, 2)

        a_tuple = named_tuple(field1=15)
        self.assertEqual(a_tuple.field1, 15)
        self.assertEqual(a_tuple.field2, 2)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""An abstraction layer describing hardware supported by Mbed OS.

Querying board database
-----------------------

For the interface to query board database, look at `mbed_tools.targets.get_board`.

Fetching target data
____________________

For the interface to extract target data from their definitions in Mbed OS,
look at `mbed_tools.targets.get_target`.

Configuration
-------------

For details about configuration of this module, look at `mbed_tools.targets.config`.
"""
from mbed_tools.targets import exceptions
from mbed_tools.targets.get_target import (
    get_target_by_name,
    get_target_by_board_type,
)
from mbed_tools.targets.get_board import (
    get_board_by_product_code,
    get_board_by_online_id,
)
from mbed_tools.targets.board import Board



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from unittest import TestCase, mock

from mbed_tools.project._internal.libraries import MbedLibReference, LibraryReferences
from tests.project.factories import make_mbed_lib_reference, patchfs


@mock.patch("mbed_tools.project._internal.git_utils.clone", autospec=True)
class TestLibraryReferences(TestCase):
    @patchfs
    def test_hydrates_top_level_library_references(self, mock_clone, fs):
        fs_root = pathlib.Path(fs, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git")
        mock_clone.side_effect = lambda url, dst_dir: dst_dir.mkdir()

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.resolve()

        mock_clone.assert_called_once_with(lib.get_git_reference().repo_url, lib.source_code_path)
        self.assertTrue(lib.is_resolved())

    @patchfs
    def test_hydrates_recursive_dependencies(self, mock_clone, fs):
        fs_root = pathlib.Path(fs, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git")
        # Create a lib reference without touching the fs at this point, we want to mock the effects of a recursive
        # reference lookup and we need to assert the reference was resolved.
        lib2 = MbedLibReference(
            reference_file=(lib.source_code_path / "lib2.lib"), source_code_path=(lib.source_code_path / "lib2")
        )
        # Here we mock the effects of a recursive reference lookup. We create a new lib reference as a side effect of
        # the first call to the mock. Then we create the src dir, thus resolving the lib, on the second call.
        mock_clone.side_effect = lambda url, dst_dir: (
            make_mbed_lib_reference(pathlib.Path(dst_dir), name=lib2.reference_file.name, ref_url="https://valid2"),
            lib2.source_code_path.mkdir(),
        )

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.resolve()

        self.assertTrue(lib.is_resolved())
        self.assertTrue(lib2.is_resolved())

    @patchfs
    @mock.patch("mbed_tools.project._internal.git_utils.checkout", autospec=True)
    @mock.patch("mbed_tools.project._internal.git_utils.init", autospec=True)
    def test_does_not_perform_deploy_if_no_git_ref_exists(self, mock_init, mock_checkout, mock_clone, fs):
        fs_root = pathlib.Path(fs, "foo")
        make_mbed_lib_reference(fs_root, ref_url="https://git", resolved=True)

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.deploy(force=False)

        mock_checkout.assert_not_called()

    @patchfs
    @mock.patch("mbed_tools.project._internal.git_utils.checkout", autospec=True)
    @mock.patch("mbed_tools.project._internal.git_utils.init", autospec=True)
    def test_performs_deploy_if_git_ref_exists(self, mock_init, mock_checkout, mock_clone, fs):
        fs_root = pathlib.Path(fs, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git#lajdhalk234", resolved=True)

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.deploy(force=False)

        mock_checkout.assert_called_once_with(mock_init.return_value, lib.get_git_reference().ref, force=False)

    @patchfs
    @mock.patch("mbed_tools.project._internal.git_utils.checkout", autospec=True)
    @mock.patch("mbed_tools.project._internal.git_utils.init", autospec=True)
    def test_resolve_does_not_perform_deploy_if_no_git_ref_exists(self, mock_init, mock_checkout, mock_clone, fs):
        fs_root = pathlib.Path(fs, "foo")
        make_mbed_lib_reference(fs_root, ref_url="https://git")
        mock_clone.side_effect = lambda url, dst_dir: dst_dir.mkdir()

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.resolve()

        mock_checkout.assert_not_called()

    @patchfs
    @mock.patch("mbed_tools.project._internal.git_utils.checkout", autospec=True)
    @mock.patch("mbed_tools.project._internal.git_utils.init", autospec=True)
    def test_resolve_performs_deploy_if_git_ref_exists(self, mock_init, mock_checkout, mock_clone, fs):
        fs_root = pathlib.Path(fs, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git#lajdhalk234")
        mock_clone.side_effect = lambda url, dst_dir: dst_dir.mkdir()

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.resolve()

        mock_checkout.assert_called_once_with(None, lib.get_git_reference().ref)

    @patchfs
    @mock.patch("mbed_tools.project._internal.git_utils.checkout", autospec=True)
    @mock.patch("mbed_tools.project._internal.git_utils.init", autospec=True)
    def test_does_not_resolve_references_in_ignore_paths(self, mock_init, mock_checkout, mock_clone, fs):
        fs_root = pathlib.Path(fs, "mbed-os")
        make_mbed_lib_reference(fs_root, ref_url="https://git#lajdhalk234")

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.resolve()

        mock_clone.assert_not_called()



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for sync_board_database.py."""
from unittest import TestCase

from ci_scripts import sync_board_database

from mbed_tools.targets.boards import Boards
from mbed_tools.targets import Board


BOARD_1 = {
    "type": "target",
    "id": "1",
    "attributes": {
        "features": {
            "mbed_enabled": ["Advanced"],
            "mbed_os_support": [
                "Mbed OS 5.10",
                "Mbed OS 5.11",
                "Mbed OS 5.12",
                "Mbed OS 5.13",
                "Mbed OS 5.14",
                "Mbed OS 5.15",
                "Mbed OS 5.8",
                "Mbed OS 5.9",
            ],
            "antenna": ["Connector", "Onboard"],
            "certification": [
                "Anatel (Brazil)",
                "AS/NZS (Australia and New Zealand)",
                "CE (Europe)",
                "FCC/CFR (USA)",
                "IC RSS (Canada)",
                "ICASA (South Africa)",
                "KCC (South Korea)",
                "MIC (Japan)",
                "NCC (Taiwan)",
                "RoHS (Europe)",
            ],
            "communication": ["Bluetooth & BLE"],
            "interface_firmware": ["DAPLink", "J-Link"],
            "target_core": ["Cortex-M4"],
            "mbed_studio_support": ["Build and run"],
        },
        "board_type": "MTB_UBLOX_NINA_B1",
        "flash_size": 512,
        "name": "u-blox NINA-B1",
        "product_code": "0455",
        "ram_size": 64,
        "target_type": "module",
        "hidden": False,
        "device_name": "nRF52832_xxAA",
    },
}

BOARD_2 = {
    "type": "target",
    "id": "2",
    "attributes": {
        "features": {
            "mbed_enabled": ["Advanced"],
            "mbed_os_support": [
                "Mbed OS 5.10",
                "Mbed OS 5.11",
                "Mbed OS 5.12",
                "Mbed OS 5.13",
                "Mbed OS 5.14",
                "Mbed OS 5.15",
                "Mbed OS 5.8",
                "Mbed OS 5.9",
            ],
            "antenna": ["Connector"],
            "certification": ["AS/NZS (Australia and New Zealand)", "CE (Europe)", "FCC/CFR (USA)", "IC RSS (Canada)"],
            "communication": ["LoRa"],
            "interface_firmware": ["DAPLink"],
            "target_core": ["Cortex-M3"],
            "mbed_studio_support": ["Build and run"],
        },
        "board_type": "MTB_MTS_XDOT",
        "flash_size": 256,
        "name": "Multitech xDOT",
        "product_code": "0453",
        "ram_size": 64,
        "target_type": "module",
        "hidden": False,
        "device_name": "STM32L151CC",
    },
}

BOARD_3 = {
    "type": "target",
    "id": "3",
    "attributes": {
        "features": {
            "mbed_enabled": ["Advanced"],
            "mbed_os_support": [
                "Mbed OS 5.10",
                "Mbed OS 5.11",
                "Mbed OS 5.12",
                "Mbed OS 5.13",
                "Mbed OS 5.14",
                "Mbed OS 5.15",
                "Mbed OS 5.7",
                "Mbed OS 5.8",
                "Mbed OS 5.9",
            ],
            "antenna": ["Connector"],
            "certification": ["CE (Europe)", "FCC/CFR (USA)", "IC RSS (Canada)", "NCC (Taiwan)"],
            "communication": ["LoRa"],
            "interface_firmware": ["DAPLink"],
            "security": ["TRNG"],
            "target_core": ["Cortex-M4"],
            "mbed_studio_support": ["Build and run"],
        },
        "board_type": "ADV_WISE_1510",
        "flash_size": 256,
        "name": "Advantech WISE-1510",
        "product_code": "0458",
        "ram_size": 64,
        "target_type": "module",
        "hidden": False,
        "device_name": "STM32L443RC",
        "slug": "advantech-wise-1510",
    },
}

BOARD_2_MODIFIED = {
    "type": "target",
    "id": "2",
    "attributes": {
        "features": {
            "mbed_enabled": ["Advanced"],
            "mbed_os_support": [
                "Mbed OS 6.2",
                "Mbed OS 5.10",
                "Mbed OS 5.11",
                "Mbed OS 5.12",
                "Mbed OS 5.13",
                "Mbed OS 5.14",
                "Mbed OS 5.15",
                "Mbed OS 5.8",
                "Mbed OS 5.9",
            ],
            "antenna": ["Connector"],
            "certification": ["AS/NZS (Australia and New Zealand)", "CE (Europe)", "FCC/CFR (USA)", "IC RSS (Canada)"],
            "communication": ["LoRa"],
            "interface_firmware": ["DAPLink"],
            "target_core": ["Cortex-M3"],
            "mbed_studio_support": ["Build and run"],
        },
        "board_type": "MTB_MTS_XDOT",
        "flash_size": 256,
        "name": "Multitech xDOT",
        "product_code": "0453",
        "ram_size": 64,
        "target_type": "module",
        "hidden": False,
        "device_name": "STM32L151CC",
    },
}


def _make_mbed_boards_for_diff(boards_a, boards_b):
    return (
        Boards(Board.from_online_board_entry(b) for b in boards_a),
        Boards(Board.from_online_board_entry(b) for b in boards_b),
    )


class TestBoardDatabasePath(TestCase):
    def test_board_database_path_exists(self):
        self.assertTrue(sync_board_database.BOARD_DATABASE_PATH.exists())


class TestDetermineBoardDatabaseUpdateResult(TestCase):
    def test_detects_added(self):
        mock_online_boards, mock_offline_boards = _make_mbed_boards_for_diff([BOARD_1, BOARD_2], [BOARD_1])
        result = sync_board_database.compare_databases(mock_offline_boards, mock_online_boards)
        self.assertEqual(len(result.boards_added), 1, "Expect one new board to be added to offline db.")
        self.assertEqual(len(result.boards_removed), 0, "Expect no boards to be removed from offline db.")

    def test_detects_removed(self):
        mock_offline_boards, mock_online_boards = _make_mbed_boards_for_diff([BOARD_1, BOARD_2], [BOARD_1])
        result = sync_board_database.compare_databases(mock_offline_boards, mock_online_boards)
        self.assertEqual(len(result.boards_added), 0, "Expect no boards to be added to offline db.")
        self.assertEqual(len(result.boards_removed), 1, "Expect one board to be removed from offline db.")

    def test_detects_modified(self):
        mock_offline_boards, mock_online_boards = _make_mbed_boards_for_diff(
            [BOARD_1, BOARD_2], [BOARD_1, BOARD_2_MODIFIED]
        )
        result = sync_board_database.compare_databases(mock_offline_boards, mock_online_boards)
        self.assertEqual(len(result.boards_modified), 1)
        self.assertEqual(len(result.boards_added), 0)
        self.assertEqual(len(result.boards_removed), 0)

    def test_detects_added_removed_and_modified(self):
        mock_online_boards, mock_offline_boards = _make_mbed_boards_for_diff(
            [BOARD_3, BOARD_2], [BOARD_1, BOARD_2_MODIFIED]
        )
        result = sync_board_database.compare_databases(mock_offline_boards, mock_online_boards)
        self.assertEqual(len(result.boards_modified), 1)
        self.assertEqual(len(result.boards_added), 1)
        self.assertEqual(len(result.boards_removed), 1)

    def test_returns_empty_when_no_change(self):
        mock_offline_boards, mock_online_boards = _make_mbed_boards_for_diff([BOARD_1], [BOARD_1])
        result = sync_board_database.compare_databases(mock_offline_boards, mock_online_boards)
        self.assertEqual(len(result.boards_added), 0, "Returns an empty targets container when no targets added")
        self.assertEqual(len(result.boards_removed), 0, "Returns an empty targets container when no targets removed.")


class TestCreateNewsFileTextFromResult(TestCase):
    def test_text_formatting_for_boards_added(self):
        mock_online_boards, mock_offline_boards = _make_mbed_boards_for_diff([BOARD_1, BOARD_2], [])
        result = sync_board_database.compare_databases(
            online_boards=mock_online_boards, offline_boards=mock_offline_boards
        )
        text = sync_board_database.create_news_file_text_from_result(result)
        self.assertEqual(
            text,
            f"Targets added: {BOARD_2['attributes']['name']}, {BOARD_1['attributes']['name']}.\n",
            "Text is formatted correctly.",
        )

    def test_text_formatting_for_boards_removed(self):
        mock_online_boards, mock_offline_boards = _make_mbed_boards_for_diff([], [BOARD_1, BOARD_2])
        result = sync_board_database.compare_databases(
            online_boards=mock_online_boards, offline_boards=mock_offline_boards
        )
        text = sync_board_database.create_news_file_text_from_result(result)
        self.assertEqual(
            text,
            f"Targets removed: {BOARD_2['attributes']['name']}, {BOARD_1['attributes']['name']}.\n",
            "Text is formatted correctly.",
        )

    def test_text_formatting_for_boards_modified(self):
        mock_online_boards, mock_offline_boards = _make_mbed_boards_for_diff([BOARD_2], [BOARD_2_MODIFIED])
        result = sync_board_database.compare_databases(
            online_boards=mock_online_boards, offline_boards=mock_offline_boards
        )
        text = sync_board_database.create_news_file_text_from_result(result)
        self.assertEqual(
            text, f"Targets modified: {BOARD_2['attributes']['name']}.\n", "Text is formatted correctly.",
        )

    def test_text_formatting_for_boards_added_removed_and_modified(self):
        mock_online_boards, mock_offline_boards = _make_mbed_boards_for_diff(
            [BOARD_3, BOARD_2], [BOARD_1, BOARD_2_MODIFIED]
        )
        result = sync_board_database.compare_databases(
            online_boards=mock_online_boards, offline_boards=mock_offline_boards
        )
        text = sync_board_database.create_news_file_text_from_result(result)
        self.assertEqual(
            text,
            f"Targets added: {BOARD_3['attributes']['name']}.\nTargets removed: {BOARD_1['attributes']['name']}.\n"
            f"Targets modified: {BOARD_2['attributes']['name']}.\n",
            "Text is formatted correctly.",
        )
